from django.template import Library

from ..settings import setting_cluster

register = Library()


@register.simple_tag(name='smart_setting')
def tag_smart_setting(global_name):
    return setting_cluster.get_setting(global_name=global_name).value


@register.simple_tag(name='smart_settings_check_changed')
def tag_smart_settings_check_changed():
    return setting_cluster.get_is_changed()



from django.http import Http404
from django.utils.translation import gettext_lazy as _

from .classes import StatisticType


class StatisticTypeViewMixin:
    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request=request, *args, **kwargs)

    def get_object(self):
        try:
            return StatisticType.get(
                slug=self.kwargs['slug']
            )
        except KeyError:
            raise Http404(
                _(message='Statistic "%s" not found.') % self.kwargs['slug']
            )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import get_content_type_kwargs_factory

from .icons import (
    icon_acl_create, icon_acl_delete, icon_acl_list, icon_acl_permissions,
    icon_global_acl_list
)
from .permissions import permission_acl_edit, permission_acl_view

link_acl_create = Link(
    icon=icon_acl_create, kwargs=get_content_type_kwargs_factory(
        variable_name='resolved_object'
    ), permission=permission_acl_edit, text=_(message='New ACL'),
    view='acls:acl_create'
)
link_acl_delete = Link(
    args='resolved_object.pk', icon=icon_acl_delete,
    permission=permission_acl_edit, tags='dangerous', text=_(message='Delete'),
    view='acls:acl_delete'
)
link_acl_list = Link(
    icon=icon_acl_list, kwargs=get_content_type_kwargs_factory(
        variable_name='resolved_object'
    ), permission=permission_acl_view, text=_(message='ACLs'), view='acls:acl_list'
)
link_acl_permissions = Link(
    args='resolved_object.pk', icon=icon_acl_permissions,
    permission=permission_acl_edit,
    text=_(message='Permissions'), view='acls:acl_permissions'
)
link_global_acl_list = Link(
    icon=icon_global_acl_list, text=_(message='Global ACLs'),
    view='acls:global_acl_list'
)



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .events import event_user_edited
from .managers import UserOptionsManager


class UserOptions(models.Model):
    """
    This model stores administrative configurations for a user accounts.
    At the moment it stores a boolean flag to restrict a user's
    ability to change their password.
    """
    user = models.OneToOneField(
        on_delete=models.CASCADE, related_name='user_options',
        to=settings.AUTH_USER_MODEL, unique=True, verbose_name=_(message='User')
    )
    block_password_change = models.BooleanField(
        default=False, verbose_name=_(
            message='Forbid this user from changing their password.'
        )
    )

    objects = UserOptionsManager()

    class Meta:
        verbose_name = _(message='User settings')
        verbose_name_plural = _(message='Users settings')

    def natural_key(self):
        return self.user.natural_key()
    natural_key.dependencies = [settings.AUTH_USER_MODEL]

    @method_event(
        event_manager_class=EventManagerSave,
        edited={
            'event': event_user_edited,
            'target': 'user'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from collections import OrderedDict

from django.forms.widgets import *  # NOQA
from django.forms.widgets import __all__ as django_forms_widgets_all
from django.forms.widgets import Media, SelectMultiple, TextInput, Widget
from django.utils.safestring import mark_safe

__all__ = django_forms_widgets_all + (
    'ColorWidget', 'DisableableSelectWidget', 'DropzoneWidget',
    'NamedMultiWidget', 'PlainWidget', 'TextAreaDiv'
)


class ColorWidget(TextInput):
    template_name = 'forms/forms/widgets/widget_color_picker.html'

    def __init__(self, attrs=None):
        attrs = attrs or {}
        attrs['type'] = 'color'
        super().__init__(attrs=attrs)


class DisableableSelectWidget(SelectMultiple):
    def create_option(self, *args, **kwargs):
        result = super().create_option(*args, **kwargs)

        # Get a keyword argument named value or the second positional argument
        # Current interface as of Django 1.11
        # def create_option(self, name, value, label, selected, index,
        # subindex=None, attrs=None):
        value = kwargs.get(
            'value', args[1]
        )

        if value in self.disabled_choices:
            result['attrs'].update(
                {'disabled': 'disabled'}
            )

        return result


class DropzoneWidget(Widget):
    template_name = 'forms/forms/widgets/dropzone.html'


class NamedMultiWidget(Widget):
    subwidgets = None
    subwidgets_order = None
    template_name = 'django/forms/widgets/multiwidget.html'

    def __init__(self, attrs=None):
        self.widgets = {}
        for name, widget in OrderedDict(self.subwidgets).items():
            self.widgets[name] = widget() if isinstance(widget, type) else widget

        if not self.subwidgets_order:
            self.subwidgets_order = list(
                self.widgets.keys()
            )

        super().__init__(attrs)

    def _get_media(self):
        "Media for a multiwidget is the combination of all media of the subwidgets"
        media = Media()
        for name, widget in self.widgets.items():
            media += widget.media
        return media
    media = property(_get_media)

    @property
    def is_hidden(self):
        return all(
            widget.is_hidden for name, widget in self.widgets.items()
        )

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        if self.is_localized:
            for widget in self.widgets:
                widget.is_localized = self.is_localized

        value = self.decompress(value)

        final_attrs = context['widget']['attrs']
        input_type = final_attrs.pop('type', None)
        id_ = final_attrs.get('id')
        subwidgets = []

        # Include new subwidgets added by subclasses after __init__.
        _subwidgets_order = self.subwidgets_order.copy()
        for widget in self.widgets.keys():
            if widget not in _subwidgets_order:
                _subwidgets_order.append(widget)

        for subwidget_entry in _subwidgets_order:
            widget_name = subwidget_entry
            widget = self.widgets[widget_name]
            if input_type is not None:
                widget.input_type = input_type
            full_widget_name = '{}_{}'.format(name, widget_name)
            try:
                widget_value = value[widget_name]
            except IndexError:
                widget_value = None
            if id_:
                widget_attrs = final_attrs.copy()
                widget_attrs['id'] = '{}_{}'.format(id_, widget_name)
            else:
                widget_attrs = final_attrs
            subwidgets.append(
                widget.get_context(
                    full_widget_name, widget_value, widget_attrs
                )['widget']
            )
        context['widget']['subwidgets'] = subwidgets
        return context

    def id_for_label(self, id_):
        if id_:
            id_ += '_{}'.format(
                list(
                    self.widgets.keys()
                )[0]
            )
        return id_

    def value_from_datadict(self, data, files, name):
        return {
            name: widget.value_from_datadict(
                data, files, name + '_%s' % name
            ) for name, widget in self.widgets.items()
        }

    def value_omitted_from_data(self, data, files, name):
        return all(
            widget.value_omitted_from_data(data, files, name + '_%s' % name)
            for name, widget in self.widgets.items()
        )

    @property
    def needs_multipart_form(self):
        return any(
            widget.needs_multipart_form for name, widget in self.widgets.items()
        )


class PlainWidget(Widget):
    """
    Class to define a form widget that effectively nulls the htmls of a
    widget and reduces the output to only it's value.
    """
    def render(self, name, value, attrs=None, renderer=None):
        return mark_safe(s='%s' % value)


class TextAreaDiv(Widget):
    """
    Class to define a form widget that simulates the behavior of a
    Textarea widget but using a div tag instead.
    """
    template_name = 'appearance/forms/widgets/textareadiv.html'



import hashlib

from django.core import serializers


class WorkflowTransitionTriggerEventBusinessLogicMixin:
    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()



from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def validation_fuction_check_path_format(setting, raw_value):
    if raw_value:
        cleaned_raw_value = raw_value.strip()

        if cleaned_raw_value:
            if cleaned_raw_value[0] == '/' or cleaned_raw_value[-1] == '/':
                raise ValidationError(
                    message=_(
                        message='The path value must not include a leading or '
                        'trailing slash.'
                    )
                )
            else:
                return cleaned_raw_value



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from .model_mixins import (
    ModelMixinUserConfirmViewBusinessLogic,
    ModelMixinUserViewModeBusinessLogic
)


class UserConfirmView(ModelMixinUserConfirmViewBusinessLogic, models.Model):
    namespace = models.CharField(
        db_index=True, max_length=200, verbose_name=_(message='Namespace')
    )
    name = models.CharField(
        db_index=True,
        help_text=_(message='Full name of the view including the namespace.'),
        max_length=200, verbose_name=_(message='Name')
    )
    user = models.ForeignKey(
        on_delete=models.CASCADE, related_name='view_confirms',
        to=settings.AUTH_USER_MODEL, verbose_name=_(message='User')
    )
    remember = models.BooleanField(
        default=False,
        help_text=_('Remember the last confirmation of the view.'),
        verbose_name=_(message='Remember')
    )

    def __str__(self):
        return self.name

    class Meta:
        unique_together = ('user', 'name')
        ordering = ('user__username', 'name')
        verbose_name = _(message='User confirm view')
        verbose_name_plural = _(message='User confirm view')


class UserViewMode(ModelMixinUserViewModeBusinessLogic, models.Model):
    namespace = models.CharField(
        db_index=True, max_length=200, verbose_name=_(message='Namespace')
    )
    name = models.CharField(
        db_index=True,
        help_text=_(message='Full name of the view including the namespace.'),
        max_length=200, verbose_name=_(message='Name')
    )
    user = models.ForeignKey(
        on_delete=models.CASCADE, related_name='view_modes',
        to=settings.AUTH_USER_MODEL, verbose_name=_(message='User')
    )
    value = models.CharField(
        db_index=True, help_text=_(
            'Stored value used to identify the display mode of the view.'
        ), max_length=5, verbose_name=_(message='Value')
    )

    class Meta:
        unique_together = ('user', 'name')
        ordering = ('user__username', 'name')
        verbose_name = _(message='User view mode')
        verbose_name_plural = _(message='User view modes')



import os
from pathlib import Path

from django.conf import settings
from django.template import loader
from django.template.base import Template
from django.template.context import Context
from django.utils.html import mark_safe
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.classes import BaseBackend
from mayan.apps.common.serialization import yaml_dump, yaml_load
from mayan.apps.task_manager.classes import Worker
from mayan.settings.literals import (
    DEFAULT_DIRECTORY_INSTALLATION, DEFAULT_OS_USERNAME,
    DEFAULT_USER_SETTINGS_FOLDER, GUNICORN_LIMIT_REQUEST_LINE,
    GUNICORN_MAX_REQUESTS, GUNICORN_REQUESTS_JITTER, GUNICORN_TIMEOUT,
    GUNICORN_WORKER_CLASS, GUNICORN_WORKERS
)


class Variable:
    def __init__(self, name, default, environment_name):
        self.name = name
        self.default = default
        self.environment_name = environment_name

    def _get_value(self):
        return os.environ.get(self.environment_name, self.default)

    def get_value(self):
        return mark_safe(
            s=self._get_value()
        )


class YAMLVariable(Variable):
    def _get_value(self):
        value = os.environ.get(self.environment_name)
        if value:
            value = yaml_load(stream=value)
        else:
            value = self.default

        return yaml_dump(
            data=value, allow_unicode=True, default_flow_style=True,
            width=999
        ).replace('...\n', '').replace('\n', '')


class PlatformTemplate(BaseBackend):
    _backend_identifier = 'name'
    _loader_module_name = 'platform_templates'
    context = {}
    context_defaults = {}
    label = None
    name = None
    settings = None
    template_name = None
    template_string = None
    variables = None

    def __str__(self):
        label = self.get_label()
        return str(label)

    def get_context(self):
        return self.context

    def get_context_defaults(self):
        return self.context_defaults

    def get_label(self):
        return self.label or self.name

    def get_settings_context(self):
        result = {}
        for setting in self.settings or ():
            if setting.value:
                result[setting.global_name] = setting.value

        return result

    def get_template_name(self):
        return self.template_name or 'platforms/{}.tmpl'.format(self.name)

    def get_variables_context(self):
        result = {}
        for variable in self.variables or ():
            result[variable.name] = variable.get_value()

        return result

    def render(self, context_string=None):
        """
        context_string allows the management command to pass context to this
        method as a JSON string
        """
        context = {}

        context.update(
            self.get_context_defaults()
        )
        context.update(
            self.get_settings_context()
        )
        context.update(
            self.get_variables_context()
        )
        # get_context goes last to server as the override.
        context.update(
            self.get_context()
        )

        if context_string:
            context.update(
                yaml_load(stream=context_string)
            )

        if self.template_string:
            template = Template(template_string=self.template_string)
            return template.render(
                context=Context(dict_=context)
            )
        else:
            return loader.render_to_string(
                context=context, template_name=self.get_template_name()
            )


class PlatformTemplateSupervisord(PlatformTemplate):
    label = _(message='Template for Supervisord.')
    name = 'supervisord'

    def __init__(self):
        self.variables = (
            Variable(
                name='GUNICORN_REQUESTS_JITTER',
                default=GUNICORN_REQUESTS_JITTER,
                environment_name='MAYAN_GUNICORN_REQUESTS_JITTER'
            ),
            Variable(
                name='GUNICORN_LIMIT_REQUEST_LINE',
                default=GUNICORN_LIMIT_REQUEST_LINE,
                environment_name='MAYAN_GUNICORN_GUNICORN_LIMIT_REQUEST_LINE'
            ),
            Variable(
                name='GUNICORN_MAX_REQUESTS',
                default=GUNICORN_MAX_REQUESTS,
                environment_name='MAYAN_GUNICORN_MAX_REQUESTS'
            ),
            Variable(
                name='GUNICORN_TIMEOUT',
                default=GUNICORN_TIMEOUT,
                environment_name='MAYAN_GUNICORN_TIMEOUT'
            ),
            Variable(
                name='GUNICORN_WORKER_CLASS',
                default=GUNICORN_WORKER_CLASS,
                environment_name='MAYAN_GUNICORN_WORKER_CLASS'
            ),
            Variable(
                name='GUNICORN_WORKERS',
                default=GUNICORN_WORKERS,
                environment_name='MAYAN_GUNICORN_WORKERS'
            ),
            Variable(
                name='GUNICORN_TIMEOUT',
                default=GUNICORN_TIMEOUT,
                environment_name='MAYAN_GUNICORN_TIMEOUT'
            ),
            Variable(
                name='INSTALLATION_PATH',
                default=DEFAULT_DIRECTORY_INSTALLATION,
                environment_name='MAYAN_INSTALLATION_PATH'
            ),
            Variable(
                name='OS_USERNAME', default=DEFAULT_OS_USERNAME,
                environment_name='MAYAN_OS_USERNAME'
            ),
            Variable(
                name='USER_SETTINGS_FOLDER',
                default=DEFAULT_USER_SETTINGS_FOLDER,
                environment_name='MAYAN_USER_SETTINGS_FOLDER'
            ),
            YAMLVariable(
                name='MEDIA_ROOT', default=settings.MEDIA_ROOT,
                environment_name='MAYAN_MEDIA_ROOT'
            )
        )

    def get_context(self):
        *_, user_settings_folder, media_root = self.variables

        return {
            'SUPERVISOR_AUTORESTART': 'true',
            'user_settings_folder': Path(
                media_root.get_value()
            ) / user_settings_folder.get_value(),
            'workers': Worker.all()
        }


class PlatformTemplateWorkerQueues(PlatformTemplate):
    label = _(message='Template showing the queues of a worker.')
    name = 'worker_queues'

    variables = (
        Variable(
            name='WORKER_NAME', default=None,
            environment_name='MAYAN_WORKER_NAME'
        ),
    )

    def get_context(self):
        worker_name = self.get_variables_context().get('WORKER_NAME')
        try:
            queues = Worker.get(name=worker_name).queues
        except KeyError:
            raise KeyError(
                'Worker name "{}" not found.'.format(worker_name)
            )

        return {
            'queues': queues, 'queue_names': sorted(
                map(lambda x: x.name, queues)
            )
        }



from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentFileAPIViewMixin
)
from mayan.apps.rest_api import generics

from .models import DetachedSignature, EmbeddedSignature
from .permissions import (
    permission_document_file_sign_detached,
    permission_document_file_sign_embedded,
    permission_document_file_signature_delete,
    permission_document_file_signature_upload,
    permission_document_file_signature_view
)
from .serializers import (
    DetachedSignatureSerializer, DetachedSignatureUploadSerializer,
    EmbeddedSignatureSerializer, SignDetachedSerializer,
    SignEmbeddedSerializer
)


class APIDocumentFileSignDetachedView(
    ParentObjectDocumentFileAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Sign a document file with a detached signature.
    """
    mayan_external_object_permission_map = {
        'POST': permission_document_file_sign_detached
    }
    lookup_url_kwarg = 'document_file_id'
    serializer_class = SignDetachedSerializer

    def get_source_queryset(self):
        return self.get_document_file_queryset()

    def object_action(self, obj, request, serializer):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        DetachedSignature.objects.sign_document_file(
            document_file=self.get_document_file(),
            key=serializer.validated_data['key'],
            passphrase=serializer.validated_data['passphrase'],
            user=self.request.user
        )


class APIDocumentFileSignEmbeddedView(
    ParentObjectDocumentFileAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Sign a document file with an embedded signature.
    """
    mayan_external_object_permission_map = {
        'POST': permission_document_file_sign_embedded
    }
    lookup_url_kwarg = 'document_file_id'
    serializer_class = SignEmbeddedSerializer

    def get_source_queryset(self):
        return self.get_document_file_queryset()

    def object_action(self, obj, request, serializer):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        EmbeddedSignature.objects.sign_document_file(
            document_file=self.get_document_file(),
            key=serializer.validated_data['key'],
            passphrase=serializer.validated_data['passphrase'],
            user=self.request.user
        )


class APIDocumentFileDetachedSignatureListView(
    ParentObjectDocumentFileAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the detached signatures of a document file.
    post: Create a detached signature for a document file.
    """
    mayan_external_object_permission_map = {
        'GET': permission_document_file_signature_view
    }
    serializer_class = DetachedSignatureSerializer

    def get_source_queryset(self):
        return DetachedSignature.objects.filter(
            document_file=self.get_document_file()
        )


class APIDocumentFileDetachedSignatureDetailView(
    ParentObjectDocumentFileAPIViewMixin, generics.RetrieveDestroyAPIView
):
    """
    delete: Delete an detached signature of the selected document.
    get: Returns the details of the selected detached signature.
    """
    mayan_external_object_permission_map = {
        'DELETE': permission_document_file_signature_delete,
        'GET': permission_document_file_signature_view,
        'POST': permission_document_file_signature_view
    }
    lookup_url_kwarg = 'detached_signature_id'
    serializer_class = DetachedSignatureSerializer

    def get_source_queryset(self):
        return DetachedSignature.objects.filter(
            document_file=self.get_document_file()
        )


class APIDocumentFileDetachedSignatureUploadView(
    ParentObjectDocumentFileAPIViewMixin, generics.CreateAPIView
):
    """
    post: Upload a detached signature file for a document file.
    """
    mayan_external_object_permission_map = {
        'POST': permission_document_file_signature_upload
    }
    lookup_url_kwarg = 'document_file_id'
    serializer_class = DetachedSignatureUploadSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'document_file': self.get_document_file()
        }

    def get_source_queryset(self):
        return self.get_document_file_queryset()


class APIDocumentFileEmbeddedSignatureListView(
    ParentObjectDocumentFileAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the embedded signatures of a document file.
    """
    serializer_class = EmbeddedSignatureSerializer
    mayan_external_object_permission_map = {
        'GET': permission_document_file_signature_view
    }

    def get_source_queryset(self):
        return EmbeddedSignature.objects.filter(
            document_file=self.get_document_file()
        )


class APIDocumentFileEmbeddedSignatureDetailView(
    ParentObjectDocumentFileAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns the details of the selected embedded signature.
    """
    mayan_external_object_permission_map = {
        'GET': permission_document_file_signature_view
    }
    lookup_url_kwarg = 'embedded_signature_id'
    serializer_class = EmbeddedSignatureSerializer

    def get_source_queryset(self):
        return EmbeddedSignature.objects.filter(
            document_file=self.get_document_file()
        )



from django.db import migrations

from mayan.apps.databases.literals import (
    DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
)


def code_document_file_size_copy(apps, schema_editor):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    for document_file in DocumentFile.objects.iterator():
        name = document_file.file.name
        document_file.file.close()

        if document_file.file.storage.exists(name=name):
            document_file_size = document_file.file.storage.size(name=name)
            if document_file_size > DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE:
                document_file_size = DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE

            document_file.size = document_file_size
            document_file.save(
                update_fields=('size',)
            )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0079_documentfile_size')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_file_size_copy,
            reverse_code=migrations.RunPython.noop
        )
    ]



from mayan.apps.icons.icons import Icon

icon_document_duplicates_list = Icon(
    driver_name='fontawesome', symbol='clone'
)
icon_duplicated_document_list = Icon(
    driver_name='fontawesome', symbol='clone'
)
icon_duplicated_document_scan = Icon(
    driver_name='fontawesome-dual', primary_symbol='clone',
    secondary_symbol='eye'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Search'), name='search'
)

event_saved_resultset_created = namespace.add_event_type(
    label=_(message='Resultset created'), name='saved_resultset_created'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.search import (
    search_model_document, search_model_document_file
)

# Document

search_model_document.add_model_field(
    field='files__file_metadata_drivers__entries__key',
    label=_(message='File metadata key')
)
search_model_document.add_model_field(
    field='files__file_metadata_drivers__entries__value',
    label=_(message='File metadata value')
)

# Document file

search_model_document_file.add_model_field(
    field='file_metadata_drivers__entries__key',
    label=_(message='File metadata key')
)
search_model_document_file.add_model_field(
    field='file_metadata_drivers__entries__value',
    label=_(message='File metadata value')
)



import json

from django.template import Library

register = Library()


@register.filter(name='json_load')
def filter_json_load(value):
    """
    Deserialize string to a Python object.
    """
    obj = json.loads(s=value)

    return obj



from mayan.apps.documents.icons import icon_document_type
from mayan.apps.icons.icons import Icon

icon_metadata = Icon(driver_name='fontawesome', symbol='database')

# Document metadata

icon_document_metadata_add = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='plus'
)
icon_document_metadata_edit = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='pen'
)
icon_document_metadata_remove = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='minus'
)
icon_document_metadata_list = Icon(
    driver_name='fontawesome', symbol='database'
)
icon_document_multiple_metadata_add = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='plus'
)
icon_document_multiple_metadata_edit = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='pen'
)
icon_document_multiple_metadata_remove = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='minus'
)

# Document type

icon_document_type_metadata_type_list = Icon(
    driver_name='fontawesome', symbol='database'
)

# Metadata type

icon_metadata_type_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='plus'
)
icon_metadata_type_delete_single = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_metadata_type_delete_multiple = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_metadata_type_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_metadata_type_list = Icon(driver_name='fontawesome', symbol='database')
icon_metadata_type_document_type_list = icon_document_type



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('document_states', '0037_populate_state_active'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workflowinstance',
            name='state_active',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='workflow_instances', to='document_states.workflowstate', verbose_name='Active state'),
        ),
    ]



from django.apps import apps


class LockBusinessLogicMixin:
    def release(self):
        """
        Release a previously held lock.
        """
        Lock = apps.get_model(app_label='lock_manager', model_name='Lock')

        try:
            lock = Lock.objects.get(
                name=self.name, creation_datetime=self.creation_datetime
            )
        except Lock.DoesNotExist:
            """
            Our lock has expired and was reassigned.
            """
        else:
            lock.delete()



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0040_alter_workflowinstance_document_and_more')
    ]

    operations = [
        migrations.AlterField(
            field=models.CharField(
                help_text='A short text describing the action. Actions '
                'are execute by alphabetical order.', max_length=255,
                verbose_name='Label'
            ),
            model_name='workflowstateaction', name='label'
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0002_workflowstate_completion')
    ]

    operations = [
        migrations.CreateModel(
            name='WorkflowRuntimeProxy',
            fields=[],
            options={
                'verbose_name': 'Workflow runtime proxy',
                'proxy': True,
                'verbose_name_plural': 'Workflow runtime proxies'
            },
            bases=('document_states.workflow',)
        ),
        migrations.CreateModel(
            name='WorkflowStateRuntimeProxy',
            fields=[],
            options={
                'verbose_name': 'Workflow state runtime proxy',
                'proxy': True,
                'verbose_name_plural': 'Workflow state runtime proxies'
            },
            bases=('document_states.workflowstate',)
        ),
        migrations.AlterModelOptions(
            name='workflow',
            options={
                'ordering': ('label',),
                'verbose_name': 'Workflow',
                'verbose_name_plural': 'Workflows'
            }
        ),
        migrations.AlterModelOptions(
            name='workflowstate',
            options={
                'ordering': ('label',),
                'verbose_name': 'Workflow state',
                'verbose_name_plural': 'Workflow states'
            }
        ),
        migrations.AlterModelOptions(
            name='workflowtransition',
            options={
                'ordering': ('label',),
                'verbose_name': 'Workflow transition',
                'verbose_name_plural': 'Workflow transitions'
            }
        )
    ]



import logging
from pathlib import Path

import sh

from django.utils.translation import gettext_lazy as _

from mayan.apps.file_metadata.classes import FileMetadataDriver
from mayan.apps.storage.utils import TemporaryDirectory

from .literals import DEFAULT_PATH_CLAMSCAN

__all__ = ('ClamScanDriver',)
logger = logging.getLogger(name=__name__)


class ClamScanDriver(FileMetadataDriver):
    argument_name_list = ('path_clamscan',)
    description = _(message='Command line anti-virus scanner.')
    label = _(message='ClamScan')
    internal_name = 'clamscan'
    mime_type_list = ('*',)

    @classmethod
    def get_argument_values_from_settings(cls):
        result = {'path_clamscan': DEFAULT_PATH_CLAMSCAN}

        setting_arguments = super().get_argument_values_from_settings()

        if setting_arguments:
            result.update(setting_arguments)

        return result

    def __init__(self, path_clamscan, **kwargs):
        super().__init__(**kwargs)

        self.command_clamscan = sh.Command(path=path_clamscan)

    def _process(self, document_file):
        if self.command_clamscan:
            with TemporaryDirectory() as temporary_folder:
                path_temporary_file = Path(
                    temporary_folder, Path(document_file.filename).name
                )

                with path_temporary_file.open(mode='xb') as temporary_fileobject:
                    document_file.save_to_file(
                        file_object=temporary_fileobject
                    )
                    temporary_fileobject.seek(0)
                    try:
                        output = self.command_clamscan(
                            temporary_fileobject.name, _ok_code=(0, 1)
                        )
                    except sh.ErrorReturnCode_2:
                        # Some error(s) occurred.
                        raise
                    else:
                        # This code path is executed is there is no exception
                        # (error code 0) or error code 1 (virus found).
                        result = {}
                        start_of_data = False
                        for line in output.split('\n'):
                            if start_of_data:
                                parts = line.split(':', 1)
                                if len(parts) > 1:
                                    key = parts[0]
                                    value = parts[1]
                                    value = value.strip()

                                    result[key] = value
                            if 'SCAN SUMMARY' in line:
                                start_of_data = True

                        return result
        else:
            logger.error(
                'clamscan binary not found, not processing document '
                'file: %s', document_file
            )



import re

from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from .classes import Setting
from .literals import NAMESPACE_VERSION_INITIAL


class SettingNamespaceMetaclass(type):
    def __call__(cls, cluster, name, **kwargs):
        try:
            instance = cluster.get_namespace(name=name)
        except KeyError:
            instance = super().__call__(
                cluster=cluster, name=name, **kwargs
            )
        finally:
            return instance


class SettingNamespace(metaclass=SettingNamespaceMetaclass):
    def __init__(
        self, cluster, name, label, migration_class=None,
        version=NAMESPACE_VERSION_INITIAL
    ):
        self.cluster = cluster
        self.migration_class = migration_class
        self.name = name
        self.label = label
        self.setting_dict = {}
        self.version = version

        self.cluster.namespace_dict[self.name] = self

    def __str__(self):
        return force_str(s=self.label)

    def do_cache_invalidate(self):
        for setting in self.setting_dict.values():
            setting.do_cache_invalidate()

    def do_post_edit_function_call(self):
        for setting in self.setting_dict.values():
            setting.do_post_edit_function_call()

    def do_migrate(self, setting):
        if self.migration_class:
            self.migration_class(namespace=self).do_migrate(setting=setting)

    def do_setting_add(self, **kwargs):
        setting = Setting(namespace=self, **kwargs)

        return setting

    def do_setting_remove(self, global_name):
        setting = self.setting_dict.get(global_name)

        self.setting_dict.pop(setting.global_name)
        self.cluster.setting_dict.pop(setting.global_name)

        return setting

    def get_configuration_file_version(self):
        return self.cluster.get_namespace_configuration(name=self.name).get(
            'version', NAMESPACE_VERSION_INITIAL
        )

    def get_setting(self, global_name):
        return self.setting_dict[global_name]

    def get_setting_list(self):
        return sorted(
            self.setting_dict.values(), key=lambda x: x.global_name
        )


SettingNamespace.verbose_name = _(message='Settings namespace')


class SettingNamespaceMigration:
    @staticmethod
    def get_method_name(setting):
        return setting.global_name.lower()

    def __init__(self, namespace):
        self.namespace = namespace

    def do_migrate(self, setting):
        if self.namespace.get_configuration_file_version() != self.namespace.version:
            setting_method_name = SettingNamespaceMigration.get_method_name(
                setting=setting
            )

            # Get methods for this setting.
            pattern = r'{}_\d{{4}}'.format(setting_method_name)
            setting_methods = re.findall(
                pattern=pattern, string='\n'.join(
                    dir(self)
                )
            )

            # Get order of execution of setting methods.
            version_list = [
                method.replace(
                    '{}_'.format(setting_method_name), ''
                ) for method in setting_methods
            ]
            try:
                start = version_list.index(
                    self.namespace.get_configuration_file_version()
                )
            except ValueError:
                start = 0

            try:
                end = version_list.index(self.namespace.version)
            except ValueError:
                end = None

            value = setting.value_raw
            for version in version_list[start:end]:
                method = getattr(
                    self, self.get_method_name_full(
                        setting=setting, version=version
                    ), None
                )
                if method:
                    value = method(value=value)

            setting.value_raw = value

    def get_method_name_full(self, setting, version):
        return '{}_{}'.format(
            self.get_method_name(setting=setting), version
        )



from mayan.apps.sources.source_backend_actions.interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestView, SourceBackendActionInterfaceTask
)
from mayan.apps.sources.source_backend_actions.mixins.callback_mixins import (
    SourceBackendActionMixinCallbackDocumentFileUpload,
    SourceBackendActionMixinCallbackDocumentUpload
)

from .arguments import argument_query_string


class SourceBackendActionMixinCallbackPostDocumentCreateUserInteractive:
    """
    Inject the user into the post document create callback kwargs.
    """
    def get_callback_kwargs_post_document_create(self, task_kwargs):
        result = super().get_callback_kwargs_post_document_create(
            task_kwargs=task_kwargs
        )

        if 'user' in task_kwargs:
            result['user_id'] = task_kwargs['user'].pk

        return result


class SourceBackendActionMixinCallbackPostDocumentFileUploadUserInteractive:
    """
    Inject the user into the post document file upload callback kwargs.
    """
    def get_callback_kwargs_post_document_file_upload(self, task_kwargs):
        result = super().get_callback_kwargs_post_document_file_upload(
            task_kwargs=task_kwargs
        )

        if 'user' in task_kwargs:
            result['user_id'] = task_kwargs['user'].pk

        return result


class SourceBackendActionMixinCallbackPostDocumentUploadQueryStringInteractive:
    """
    Inject the query string into the post document create callback kwargs.
    """
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                query_string = argument_query_string

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['query_string'] = self.context['query_string']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['query_string'] = self.action.get_query_string(
                    request=self.context['request']
                )

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                query_string = argument_query_string

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['query_string'] = self.context['query_string']

        class View(SourceBackendActionInterfaceRequestView):
            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['query_string'] = self.action.get_query_string(
                    request=self.context['request']
                )

    def get_callback_kwargs_post_document_create(self, task_kwargs):
        result = super().get_callback_kwargs_post_document_create(
            task_kwargs=task_kwargs
        )

        result['query_string'] = task_kwargs['query_string']

        return result

    def get_query_string(self, request):
        query_string = ''

        query_dict = request.GET.copy()
        query_dict.update(request.POST)

        # Convert into a string. Make sure it is a `QueryDict` object from a
        # request and not just a simple dictionary.
        if hasattr(query_dict, 'urlencode'):
            query_string = query_dict.urlencode()

        return query_string

    def get_task_kwargs(self, query_string, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['query_string'] = query_string

        return result


class SourceBackendActionMixinCallbackDocumentUploadInteractive(
    SourceBackendActionMixinCallbackPostDocumentFileUploadUserInteractive,
    SourceBackendActionMixinCallbackPostDocumentUploadQueryStringInteractive,
    SourceBackendActionMixinCallbackPostDocumentCreateUserInteractive,
    SourceBackendActionMixinCallbackDocumentUpload
):
    """
    Interactive callback mixin for document uploads.
    """


class SourceBackendActionMixinCallbackDocumentFileUploadInteractive(
    SourceBackendActionMixinCallbackPostDocumentFileUploadUserInteractive,
    SourceBackendActionMixinCallbackDocumentFileUpload
):
    """
    Interactive callback mixin for document file uploads.
    """



from mayan.apps.icons.icons import Icon

# App

icon_statistics = Icon(driver_name='fontawesome', symbol='chart-line')

# Statistic

icon_statistic_detail = Icon(driver_name='fontawesome', symbol='eye')
icon_statistic_queue = Icon(driver_name='fontawesome', symbol='cog')

# Statistic namespace

icon_statistic_namespace_detail = Icon(
    driver_name='fontawesome', symbol='chart-line'
)
icon_statistic_namespace_list = Icon(
    driver_name='fontawesome', symbol='chart-line'
)



DEFAULT_FILE_METADATA_AUTO_PROCESS = True
DEFAULT_FILE_METADATA_DRIVERS_ARGUMENTS = {}

ERROR_LOG_DOMAIN_NAME = 'file_metadata'

LOCK_EXPIRE = 60 * 10  # Adjust to worst case scenario

RESULT_SEPARATOR = '__'



from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.classes import BinaryDependency

from .backends.literals import DEFAULT_FILE_PATH, DEFAULT_MIMETYPE_PATH

BinaryDependency(
    label='File::MimeInfo', help_text=_(
        message='This module can be used to determine the MIME type of a '
        'file. It tries to implement the freedesktop specification for a '
        'shared MIME database.'
    ), module=__name__, name='libfile-mimeinfo-perl',
    path=DEFAULT_MIMETYPE_PATH
)
BinaryDependency(
    label='file', help_text=_(
        message='determine file type using content tests'
    ), module=__name__, name='file', path=DEFAULT_FILE_PATH
)



from mayan.apps.documents.tasks import task_document_file_create

from .document_file_action_mixins import (
    SourceBackendActionMixinDocumentFileActionInteractive
)
from .document_file_comment_mixins import (
    SourceBackendActionMixinDocumentFileCommentInteractive
)
from .document_file_filename_mixins import (
    SourceBackendActionMixinDocumentFileFilenameInteractive
)
from .user_mixins import SourceBackendActionMixinUserInteractive


class SourceBackendActionMixinDocumentFileUploadInteractiveBase:
    def _background_task(self, **kwargs):
        result = super()._background_task(**kwargs)
        document = result['document']

        for server_upload_entry in result['server_upload_entry_list']:
            document_file_task_kwargs = self.get_document_file_task_kwargs(
                server_upload_entry=server_upload_entry, **kwargs
            )

            document_file_task_kwargs['shared_uploaded_file_id'] = server_upload_entry[
                'shared_uploaded_file_id'
            ]
            document_file_task_kwargs['document_id'] = document.pk

            task_document_file_create.apply_async(
                kwargs=document_file_task_kwargs
            )

    def get_document_file_task_kwargs(self, **kwargs):
        result = super().get_document_file_task_kwargs(**kwargs)

        result['is_document_upload_sequence'] = True

        return result

    def get_task_kwargs(self, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        return result


class SourceBackendActionMixinDocumentFileUploadInteractive(
    SourceBackendActionMixinDocumentFileActionInteractive,
    SourceBackendActionMixinDocumentFileCommentInteractive,
    SourceBackendActionMixinDocumentFileFilenameInteractive,
    SourceBackendActionMixinUserInteractive,
    SourceBackendActionMixinDocumentFileUploadInteractiveBase
):
    """
    Mixin for a complete action that uploads document files.
    """



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from ..models import WorkflowState, WorkflowStateAction


class WorkflowTemplateStateSerializer(serializers.HyperlinkedModelSerializer):
    actions_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Actions URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_state_id'
            }
        ), view_name='rest_api:workflow-template-state-action-list'
    )
    documents_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Documents URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_state_id'
            }
        ), view_name='rest_api:workflow-template-state-document-list'
    )
    escalations_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Escalations URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_state_id'
            }
        ), view_name='rest_api:workflow-template-state-escalation-list'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_state_id'
            }
        ), view_name='rest_api:workflow-template-state-detail'
    )
    workflow_template_id = serializers.IntegerField(
        label=_(message='Workflow template ID'), read_only=True, source='workflow_id'
    )
    workflow_template_url = serializers.HyperlinkedIdentityField(
        label=_(message='Workflow template URL'), lookup_field='workflow_id',
        lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-detail'
    )

    class Meta:
        fields = (
            'actions_url', 'completion', 'documents_url', 'escalations_url',
            'final', 'id', 'initial', 'label', 'url', 'workflow_template_id',
            'workflow_template_url'
        )
        model = WorkflowState
        read_only_fields = (
            'documents_url', 'escalations_url', 'id', 'url',
            'workflow_template_id', 'workflow_template_url'
        )


class WorkflowTemplateStateActionSerializer(
    serializers.HyperlinkedModelSerializer
):
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'state__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'state_id',
                'lookup_url_kwarg': 'workflow_template_state_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_state_action_id'
            }
        ), view_name='rest_api:workflow-template-state-action-detail'
    )
    workflow_template_state_id = serializers.IntegerField(
        label=_(message='Workflow template state ID'), read_only=True,
        source='state_id'
    )
    workflow_template_state_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow template state URL'), view_kwargs=(
            {
                'lookup_field': 'state__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'state_id',
                'lookup_url_kwarg': 'workflow_template_state_id'
            }
        ), view_name='rest_api:workflow-template-state-detail'
    )

    class Meta:
        fields = (
            'backend_data', 'backend_path', 'condition', 'enabled', 'id',
            'label', 'url', 'when', 'workflow_template_state_id',
            'workflow_template_state_url'
        )
        model = WorkflowStateAction
        read_only_fields = (
            'id', 'url', 'workflow_template_state_id',
            'workflow_template_state_url'
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_SOURCES_BACKEND_ARGUMENTS, DEFAULT_SOURCES_CACHE_MAXIMUM_SIZE,
    DEFAULT_SOURCES_CACHE_STORAGE_BACKEND,
    DEFAULT_SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS
)
from .setting_callbacks import callback_update_source_cache_size
from .setting_migrations import SourcesSettingMigration

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Sources'), migration_class=SourcesSettingMigration,
    name='sources', version='0003'
)

setting_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_SOURCES_BACKEND_ARGUMENTS,
    global_name='SOURCES_BACKEND_ARGUMENTS', help_text=_(
        message='Arguments to use when creating source backends.'
    )
)

setting_source_cache_maximum_size = setting_namespace.do_setting_add(
    default=DEFAULT_SOURCES_CACHE_MAXIMUM_SIZE,
    global_name='SOURCES_CACHE_MAXIMUM_SIZE',
    help_text=_(
        message='The threshold at which the SOURCES_CACHE_STORAGE_BACKEND '
        'will start deleting the oldest source cache files. Specify the size '
        'in bytes.'
    ), post_edit_function=callback_update_source_cache_size
)
setting_source_cache_storage_backend = setting_namespace.do_setting_add(
    global_name='SOURCES_CACHE_STORAGE_BACKEND',
    default=DEFAULT_SOURCES_CACHE_STORAGE_BACKEND, help_text=_(
        message='Path to the Storage subclass used to store cached source '
        'image files.'
    )
)
setting_source_cache_storage_backend_arguments = setting_namespace.do_setting_add(
    global_name='SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS',
    default=DEFAULT_SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS, help_text=_(
        message='Arguments to pass to the '
        'SOURCES_SOURCE_CACHE_STORAGE_BACKEND.'
    )
)



from ..interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)

from .arguments import argument_document_language


class SourceBackendActionMixinLanguageBase:
    def get_document_task_kwargs(self, language=None, **kwargs):
        result = super().get_document_task_kwargs(**kwargs)

        result['language'] = language

        return result

    def get_task_kwargs(self, language=None, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['language'] = language

        return result


class SourceBackendActionMixinLanguageInteractive(
    SourceBackendActionMixinLanguageBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                language = argument_document_language

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['language'] = self.context['language']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                language = argument_document_language

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['language'] = self.context['language']

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                language = argument_document_language

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['language'] = self.context['language']

        class View(SourceBackendActionInterfaceRequestViewForm):
            def process_interface_context(self):
                super().process_interface_context()

                form_document = self.context['forms']['document_form']
                form_document_cleaned_data = form_document.cleaned_data

                self.action_kwargs['language'] = form_document_cleaned_data.get(
                    'language', None
                )


class SourceBackendActionMixinLanguageInteractiveNot(
    SourceBackendActionMixinLanguageBase
):
    class Interface:
        class Task(SourceBackendActionInterfaceTask):
            def process_interface_context(self):
                super().process_interface_context()

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['language'] = source_backend_instance.get_document_language()



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import (
    icon_comment_add, icon_comment_delete, icon_comment_edit,
    icon_comments_for_document
)
from .permissions import (
    permission_document_comment_create, permission_document_comment_delete,
    permission_document_comment_edit, permission_document_comment_view
)

link_comment_add = Link(
    args='object.pk', icon=icon_comment_add,
    permission=permission_document_comment_create, text=_(message='Add comment'),
    view='comments:comment_add'
)
link_comment_delete = Link(
    args='object.pk', icon=icon_comment_delete,
    permission=permission_document_comment_delete, tags='dangerous',
    text=_(message='Delete'), view='comments:comment_delete'
)
link_comment_edit = Link(
    args='object.pk', icon=icon_comment_edit,
    permission=permission_document_comment_edit,
    text=_(message='Edit'), view='comments:comment_edit'
)
link_comments_for_document = Link(
    args='resolved_object.pk', icon=icon_comments_for_document,
    permission=permission_document_comment_view, text=_(message='Comments'),
    view='comments:comments_for_document'
)



from mayan.apps.icons.icons import Icon

# Document version

icon_document_version_ocr_content_delete_single = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='times'
)
icon_document_version_ocr_content_delete_multiple = icon_document_version_ocr_content_delete_single
icon_document_version_ocr_content_detail = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_ocr_submit_single = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-right'
)
icon_document_version_ocr_submit_multiple = icon_document_version_ocr_submit_single

# Document version page

icon_document_version_page_ocr_content_detail = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_page_ocr_content_edit = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_ocr_content_download = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-down'
)

# Document type

icon_document_type_ocr_settings = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_type_ocr_submit = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-right'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.platforms.platform_templates import PlatformTemplate, Variable
from mayan.settings.literals import (
    DEFAULT_DATABASE_NAME, DEFAULT_DATABASE_PASSWORD, DEFAULT_DATABASE_USER,
    DOCKER_CLI_IMAGE_VERSION, DOCKER_DIND_IMAGE_VERSION,
    DOCKER_LINUX_IMAGE_NAME, DOCKER_LINUX_IMAGE_TAG,
    DOCKER_MYSQL_IMAGE_NAME, DOCKER_MYSQL_IMAGE_TAG,
    DOCKER_POSTGRESQL_IMAGE_NAME, DOCKER_POSTGRESQL_IMAGE_TAG,
    GITLAB_CI_BRANCH_BUILDS_DOCKER, GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION,
    GITLAB_CI_BRANCH_BUILDS_PYTHON, GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO,
    GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING, GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR,
    GITLAB_CI_BRANCH_RELEASES_ALL_MINOR,
    GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR,
    GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR,
    GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION,
    GITLAB_CI_BRANCH_RELEASES_NIGHTLY, GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR,
    GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR, GITLAB_CI_BRANCH_RELEASES_STAGING,
    GITLAB_CI_BRANCH_RELEASES_TESTING, GITLAB_CI_BRANCH_TESTS_ALL,
    GITLAB_CI_BRANCH_TESTS_DOCKER, GITLAB_CI_BRANCH_TESTS_PYTHON_ALL,
    GITLAB_CI_BRANCH_TESTS_PYTHON_BASE, GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE,
    GITLAB_CI_BRANCH_TRY_STAGING, LINUX_PACKAGES_ALPINE_BUILD,
    LINUX_PACKAGES_ALPINE_PYTHON, LINUX_PACKAGES_DEBIAN_BASE,
    LINUX_PACKAGES_DEBIAN_BUILD, LINUX_PACKAGES_DEBIAN_DOCUMENTATION,
    LINUX_PACKAGES_DEBIAN_POSTGRESQL, LINUX_PACKAGES_DEBIAN_PYTHON,
    LINUX_PACKAGES_DEBIAN_TEST
)


class PlatformTemplateGitLabCI(PlatformTemplate):
    label = _(message='Template that generates a GitLab CI config file.')
    name = 'gitlab-ci'
    template_name = 'platform/gitlab/gitlab-ci.tmpl'

    def __init__(self):
        self.variables = (
            Variable(
                name='DEFAULT_DATABASE_NAME',
                default=DEFAULT_DATABASE_NAME,
                environment_name='MAYAN_DEFAULT_DATABASE_NAME'
            ),
            Variable(
                name='DEFAULT_DATABASE_PASSWORD',
                default=DEFAULT_DATABASE_PASSWORD,
                environment_name='MAYAN_DEFAULT_DATABASE_PASSWORD'
            ),
            Variable(
                name='DEFAULT_DATABASE_USER',
                default=DEFAULT_DATABASE_USER,
                environment_name='MAYAN_DEFAULT_DATABASE_USER'
            ),
            Variable(
                name='DOCKER_CLI_IMAGE_VERSION',
                default=DOCKER_CLI_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_CLI_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_DIND_IMAGE_VERSION',
                default=DOCKER_DIND_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_DIND_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_LINUX_IMAGE_NAME',
                default=DOCKER_LINUX_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_LINUX_IMAGE_TAG',
                default=DOCKER_LINUX_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_MYSQL_IMAGE_NAME',
                default=DOCKER_MYSQL_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_MYSQL_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_MYSQL_IMAGE_TAG',
                default=DOCKER_MYSQL_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_MYSQL_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_NAME',
                default=DOCKER_POSTGRESQL_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_TAG',
                default=DOCKER_POSTGRESQL_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_TAG'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_BUILDS_DOCKER',
                default=GITLAB_CI_BRANCH_BUILDS_DOCKER,
                environment_name='MAYAN_GITLAB_CI_BRANCH_BUILDS_DOCKER'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION',
                default=GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION,
                environment_name='MAYAN_GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_BUILDS_PYTHON',
                default=GITLAB_CI_BRANCH_BUILDS_PYTHON,
                environment_name='MAYAN_GITLAB_CI_BRANCH_BUILDS_PYTHON'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO',
                default=GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO,
                environment_name='MAYAN_GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING',
                default=GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR',
                default=GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_ALL_MINOR',
                default=GITLAB_CI_BRANCH_RELEASES_ALL_MINOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_ALL_MINOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR',
                default=GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR',
                default=GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION',
                default=GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_NIGHTLY',
                default=GITLAB_CI_BRANCH_RELEASES_NIGHTLY,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_NIGHTLY'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR',
                default=GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR',
                default=GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_STAGING',
                default=GITLAB_CI_BRANCH_RELEASES_STAGING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_STAGING'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_TESTING',
                default=GITLAB_CI_BRANCH_RELEASES_TESTING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_TESTING'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_ALL',
                default=GITLAB_CI_BRANCH_TESTS_ALL,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_ALL'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_DOCKER',
                default=GITLAB_CI_BRANCH_TESTS_DOCKER,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_DOCKER'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_PYTHON_ALL',
                default=GITLAB_CI_BRANCH_TESTS_PYTHON_ALL,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_PYTHON_ALL'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_PYTHON_BASE',
                default=GITLAB_CI_BRANCH_TESTS_PYTHON_BASE,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_PYTHON_BASE'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE',
                default=GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TRY_STAGING',
                default=GITLAB_CI_BRANCH_TRY_STAGING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TRY_STAGING'
            )
        )

    def get_context(self):
        return {
            'LINUX_PACKAGES_ALPINE_BUILD': LINUX_PACKAGES_ALPINE_BUILD,
            'LINUX_PACKAGES_ALPINE_PYTHON': LINUX_PACKAGES_ALPINE_PYTHON,
            'LINUX_PACKAGES_DEBIAN_BASE': LINUX_PACKAGES_DEBIAN_BASE,
            'LINUX_PACKAGES_DEBIAN_BUILD': LINUX_PACKAGES_DEBIAN_BUILD,
            'LINUX_PACKAGES_DEBIAN_DOCUMENTATION': LINUX_PACKAGES_DEBIAN_DOCUMENTATION,
            'LINUX_PACKAGES_DEBIAN_POSTGRESQL': LINUX_PACKAGES_DEBIAN_POSTGRESQL,
            'LINUX_PACKAGES_DEBIAN_PYTHON': LINUX_PACKAGES_DEBIAN_PYTHON,
            'LINUX_PACKAGES_DEBIAN_TEST': LINUX_PACKAGES_DEBIAN_TEST
        }



import logging

from django.utils.translation import gettext_lazy as _, ngettext

from mayan.apps.views.exceptions import ActionError
from mayan.apps.views.generics import MultipleObjectConfirmActionView

from ..icons import (
    icon_document_favorite_add_multiple, icon_document_favorite_list,
    icon_document_favorite_remove_multiple
)
from ..models.document_models import Document
from ..models.favorite_document_models import FavoriteDocument
from ..permissions import permission_document_view
from ..settings import setting_favorite_count

from .document_views import DocumentListView

logger = logging.getLogger(name=__name__)


class FavoriteDocumentListView(DocumentListView):
    view_icon = icon_document_favorite_list

    def get_document_queryset(self):
        return FavoriteDocument.valid.get_for_user(user=self.request.user)

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_icon': icon_document_favorite_list,
                'no_results_text': _(
                    message='Favorited documents will be listed in this view. '
                    'Up to %(count)d documents can be favorited per user. '
                ) % {'count': setting_favorite_count.value},
                'no_results_title': _(message='There are no favorited documents.'),
                'title': _(message='Favorites')
            }
        )
        return context


class FavoriteAddView(MultipleObjectConfirmActionView):
    object_permission = permission_document_view
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message = _(
        message='%(count)d document added to favorites.'
    )
    success_message_plural = _(
        message='%(count)d documents added to favorites.'
    )
    view_icon = icon_document_favorite_add_multiple

    def get_extra_context(self):
        context = {
            'title': ngettext(
                singular='Add the selected document to favorites?',
                plural='Add the selected documents to favorites?',
                number=self.object_list.count()
            )
        }

        if self.object_list.count() == 1:
            context['object'] = self.object_list.first()

        return context

    def object_action(self, form, instance):
        FavoriteDocument.valid.add_for_user(
            document=instance, user=self.request.user
        )


class FavoriteRemoveView(MultipleObjectConfirmActionView):
    error_message = _(message='Document "%(instance)s" is not in favorites.')
    object_permission = permission_document_view
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message = _(
        message='%(count)d document removed from favorites.'
    )
    success_message_plural = _(
        message='%(count)d documents removed from favorites.'
    )
    view_icon = icon_document_favorite_remove_multiple

    def get_extra_context(self):
        context = {
            'title': ngettext(
                singular='Remove the selected document from favorites?',
                plural='Remove the selected documents from favorites?',
                number=self.object_list.count()
            )
        }

        if self.object_list.count() == 1:
            context['object'] = self.object_list.first()

        return context

    def object_action(self, form, instance):
        try:
            FavoriteDocument.valid.remove_for_user(
                document=instance, user=self.request.user
            )
        except FavoriteDocument.DoesNotExist:
            raise ActionError



from django.urls import re_path

from .api_views.event_api_views import (
    APIEventListView, APIObjectEventListView
)
from .api_views.event_type_api_views import (
    APIEventTypeListView, APIEventTypeNamespaceDetailView,
    APIEventTypeNamespaceEventTypeListView, APIEventTypeNamespaceListView,
)
from .api_views.notification_api_views import (
    APINotificationDetailView, APINotificationListView
)
from .views.clear_views import (
    EventListClearView, ObjectEventClearView, VerbEventClearView
)
from .views.event_views import (
    EventListView, ObjectEventListView, VerbEventListView
)
from .views.export_views import (
    EventListExportView, ObjectEventExportView, VerbEventExportView
)
from .views.notification_views import (
    NotificationDeleteView, NotificationListView, NotificationMarkReadView,
    NotificationMarkReadAllView
)
from .views.subscription_views import (
    EventTypeSubscriptionListView, ObjectEventTypeSubscriptionListView,
    UserObjectSubscriptionList
)

urlpatterns_events = [
    re_path(
        route=r'^events/$', name='event_list', view=EventListView.as_view()
    ),
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/$',
        name='object_event_list', view=ObjectEventListView.as_view()
    ),
    re_path(
        route=r'^verbs/(?P<verb>[\w\-\.]+)/events/$', name='verb_event_list',
        view=VerbEventListView.as_view()
    )
]

urlpatterns_events_clear = [
    re_path(
        route=r'^events/clear/$', name='event_list_clear',
        view=EventListClearView.as_view()
    ),
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/clear/$',
        name='object_event_list_clear', view=ObjectEventClearView.as_view()
    ),
    re_path(
        route=r'^verbs/(?P<verb>[\w\-\.]+)/events/clear/$',
        name='verb_event_list_clear', view=VerbEventClearView.as_view()
    )
]

urlpatterns_events_export = [
    re_path(
        route=r'^events/export/$', name='event_list_export',
        view=EventListExportView.as_view()
    ),
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/export/$',
        name='object_event_list_export', view=ObjectEventExportView.as_view()
    ),
    re_path(
        route=r'^verbs/(?P<verb>[\w\-\.]+)/events/export/$',
        name='verb_event_list_export', view=VerbEventExportView.as_view(),
    )
]

urlpatterns_notification = [
    re_path(
        route=r'^user/notifications/$', name='user_notifications_list',
        view=NotificationListView.as_view()
    ),
    re_path(
        route=r'^user/notifications/multiple/delete/$',
        name='notification_delete_multiple',
        view=NotificationDeleteView.as_view()
    ),
    re_path(
        route=r'^user/notifications/(?P<notification_id>\d+)/delete/$',
        name='notification_delete_single',
        view=NotificationDeleteView.as_view()
    ),
    re_path(
        route=r'^user/notifications/(?P<notification_id>\d+)/mark_read/$',
        name='notification_mark_read', view=NotificationMarkReadView.as_view()
    ),
    re_path(
        route=r'^user/notifications/all/mark_read/$',
        name='notification_mark_read_all',
        view=NotificationMarkReadAllView.as_view()
    )
]

urlpatterns_subscriptions = [
    re_path(
        route=r'^user/event_types/subscriptions/$',
        name='event_type_user_subscription_list',
        view=EventTypeSubscriptionListView.as_view()
    ),
    re_path(
        route=r'^user/object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/subscriptions/$',
        name='object_event_type_user_subscription_list',
        view=ObjectEventTypeSubscriptionListView.as_view()
    ),
    re_path(
        route=r'^user/object/subscriptions/$',
        name='user_object_subscription_list',
        view=UserObjectSubscriptionList.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_events)
urlpatterns.extend(urlpatterns_events_clear)
urlpatterns.extend(urlpatterns_events_export)
urlpatterns.extend(urlpatterns_notification)
urlpatterns.extend(urlpatterns_subscriptions)

api_urls_event_types = [
    re_path(
        route=r'^event_type_namespaces/(?P<name>[-\w]+)/$',
        name='event-type-namespace-detail',
        view=APIEventTypeNamespaceDetailView.as_view()
    ),
    re_path(
        route=r'^event_type_namespaces/(?P<name>[-\w]+)/event_types/$',
        name='event-type-namespace-event-type-list',
        view=APIEventTypeNamespaceEventTypeListView.as_view()
    ),
    re_path(
        route=r'^event_type_namespaces/$',
        name='event-type-namespace-list',
        view=APIEventTypeNamespaceListView.as_view()
    ),
    re_path(
        route=r'^event_types/$', name='event-type-list',
        view=APIEventTypeListView.as_view()
    )
]

api_urls_events = [
    re_path(
        route=r'^events/$', name='event-list', view=APIEventListView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/$',
        name='object-event-list', view=APIObjectEventListView.as_view()
    )
]

api_urls_notifications = [
    re_path(
        route=r'^notifications/$', name='notification-list',
        view=APINotificationListView.as_view()
    ),
    re_path(
        route=r'^notifications/(?P<notification_id>[0-9]+)/$',
        name='notification-detail', view=APINotificationDetailView.as_view()
    )
]

api_urls = []
api_urls.extend(api_urls_event_types)
api_urls.extend(api_urls_events)
api_urls.extend(api_urls_notifications)



import logging

from django.db import models
from django.db.models.base import ModelBase
from django.utils.translation import gettext_lazy as _

from .class_mixins import MixinConditionTemplate

logger = logging.getLogger(name=__name__)


class ExtraDataModelMixin:
    def __init__(self, *args, **kwargs):
        _instance_extra_data = kwargs.pop(
            '_instance_extra_data', {}
        )
        result = super().__init__(*args, **kwargs)
        for key, value in _instance_extra_data.items():
            setattr(self, key, value)

        return result


class MetaClassModelMixinConditionField(ModelBase):
    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)

        field = cls._meta.get_field('condition')
        field.help_text = cls._condition_help_text

        return cls


class ModelMixinConditionField(
    MixinConditionTemplate, models.Model,
    metaclass=MetaClassModelMixinConditionField
):
    _condition_help_text = _(
        message='The condition that will determine if this object '
        'is executed or not. Conditions that do not return any value, '
        'that return the Python logical None, or an empty string (\'\') '
        'are considered to be logical false, any other value is '
        'considered to be the logical true.'
    )

    condition = models.TextField(
        blank=True, verbose_name=_(message='Condition')
    )

    class Meta:
        abstract = True


class ValueChangeModelMixin:
    @classmethod
    def from_db(cls, db, field_names, values):
        new = super().from_db(db=db, field_names=field_names, values=values)
        new._values_previous = dict(
            zip(field_names, values)
        )
        return new

    def __init__(self, *args, **kwargs):
        self._values_previous = kwargs
        super().__init__(*args, **kwargs)

    def _get_field_previous_value(self, field):
        return self._values_previous[field]

    def _has_field_changed(self, field):
        if self._state.adding:
            if getattr(self, field) != self._values_previous[field]:
                return True

        return False



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.menus import menu_topbar

from .links import link_ajax_refresh


class AppearanceApp(MayanAppConfig):
    app_namespace = 'appearance'
    app_url = 'appearance'
    has_javascript_translations = True
    has_static_media = True
    has_tests = False
    name = 'mayan.apps.appearance'
    static_media_ignore_patterns = (
        'AUTHORS*', 'CHANGE*', 'CONTRIBUT*', 'CODE_OF_CONDUCT*', 'Grunt*',
        'MAINTAIN*', 'README*', '*.less', '*.md', '*.nupkg', '*.nuspec',
        '*.scss*', '*.sh', '*tests*', 'bower*', 'composer.json*',
        'demo*', 'grunt*', 'gulp*', 'install', 'less', 'package.json*',
        'package-lock*', 'test', 'tests', 'variable*', '*.xcf',
        'appearance/node_modules/@fancyapps/fancybox/docs/*',
        'appearance/node_modules/@fancyapps/fancybox/src/*',
        'appearance/node_modules/jquery/src/*',
        'appearance/node_modules/jquery-form/_config.yml',
        'appearance/node_modules/jquery-form/form.jquery.json',
        'appearance/node_modules/jquery-form/docs/*',
        'appearance/node_modules/jquery-form/src/*',
        'appearance/node_modules/select2/src/*',
        'appearance/node_modules/toastr/karma.conf.js',
        'appearance/node_modules/toastr/toastr.js',
        'appearance/node_modules/toastr/toastr-icon.png',
        'appearance/node_modules/toastr/nuget/*'
    )
    verbose_name = _(message='Appearance')

    def ready(self):
        super().ready()

        menu_topbar.bind_links(
            links=(link_ajax_refresh,)
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import forms

from .models import Cache, CachePartition


class CacheDetailForm(forms.DetailForm):
    fieldsets = (
        (
            _(message='Identification'), {
                'fields': ('label',)
            }
        ), (
            _(message='Storage'), {
                'fields': ('defined_storage_name',)
            }
        ), (
            _(message='Configuration'), {
                'fields': (
                    'get_maximum_size_display', 'get_total_size_display'
                )
            }
        ), (
            _(message='Objects'), {
                'fields': ('get_partition_count', 'get_partition_file_count')
            }
        )
    )

    class Meta:
        extra_fields = (
            {'field': 'label'},
            {'field': 'get_maximum_size_display'},
            {'field': 'get_total_size_display'},
            {'field': 'get_partition_count'},
            {'field': 'get_partition_file_count'}
        )
        fields = ('defined_storage_name',)
        model = Cache


class CachePartitionDetailForm(forms.DetailForm):
    class Meta:
        fields = ('cache', 'name')
        model = CachePartition



from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import (
    ExternalContentTypeObjectAPIViewMixin
)

from .permissions import (
    permission_error_log_entry_delete, permission_error_log_entry_view
)
from .serializers import ErrorLogPartitionEntrySerializer


class APIErrorLogPartitionEntryListView(
    ExternalContentTypeObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the object's error log entries
    """
    mayan_external_object_permission_map = {
        'GET': permission_error_log_entry_view
    }
    serializer_class = ErrorLogPartitionEntrySerializer

    def get_source_queryset(self):
        return self.get_external_object().error_log.all()


class APIErrorLogPartitionEntryDetailView(
    ExternalContentTypeObjectAPIViewMixin, generics.RetrieveDestroyAPIView
):
    """
    delete: Delete the selected error log entry.
    get: Returns the details of the selected error log entry.
    """
    lookup_url_kwarg = 'error_log_partition_entry_id'
    mayan_external_object_permission_map = {
        'DELETE': permission_error_log_entry_delete,
        'GET': permission_error_log_entry_view
    }
    serializer_class = ErrorLogPartitionEntrySerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.get_external_object().error_log.all()



import math

from django.template import Library

register = Library()


def _process_value(value):
    if str(value) == value:
        try:
            value = float(value)
        except ValueError:
            value = int(value)

    return value


@register.filter(name='math_add')
def filter_math_add(value, argument):
    """
    Mathematical addition.
    """

    addends = _process_value(value=value), _process_value(value=argument)
    result = addends[0] + addends[1]
    return result


@register.filter(name='math_absolute')
def filter_math_absolute(value):
    """
    Mathematical absolute.
    """

    value = _process_value(value=value)
    result = abs(value)
    return result


@register.filter(name='math_divide')
def filter_math_divide(value, argument):
    """
    Mathematical division.
    """

    dividend = _process_value(value=value)
    divisor = _process_value(value=argument)
    quotient = dividend / divisor
    return quotient


@register.filter(name='math_exponentiate')
def fitler_math_exponentiate(value, argument):
    """
    Mathematical exponentiation.
    """

    base = _process_value(value=value)
    exponent = _process_value(value=argument)
    power = base ** exponent
    return power


@register.filter(name='math_floor_divide')
def filter_math_floor_divide(value, argument):
    """
    Mathematical floor division.
    """

    dividend = _process_value(value=value)
    divisor = _process_value(value=argument)
    quotient = dividend // divisor
    return quotient


@register.filter(name='math_modulo')
def filter_math_modulo(value, argument):
    """
    Mathematical modulo.
    """

    dividend = _process_value(value=value)
    divisor = _process_value(value=argument)
    modulus = dividend % divisor
    return modulus


@register.filter(name='math_multiply')
def filter_math_multiply(value, argument):
    """
    Mathematical multiplication.
    """

    multiplicand = _process_value(value=value)
    multiplier = _process_value(value=argument)
    product = multiplicand * multiplier
    return product


@register.filter(name='math_square_root')
def filter_math_square_root(value):
    """
    Mathematical square root.
    """

    radicand = _process_value(value=value)
    square_root = math.sqrt(radicand)
    return square_root


@register.filter(name='math_substract')
def filter_math_substract(value, argument):
    """
    Mathematical subtraction.
    """

    minuend = _process_value(value=value)
    subtrahend = _process_value(value=argument)
    difference = minuend - subtrahend
    return difference



import json
import re

import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.deconstruct import deconstructible
from django.utils.functional import SimpleLazyObject
from django.utils.translation import gettext_lazy as _


def _lazy_re_compile(regex, flags=0):
    """Lazily compile a regex with flags."""
    def _compile():
        # Compile the regex if it was not passed pre-compiled.
        if isinstance(regex, str):
            return re.compile(regex, flags)
        else:
            assert not flags, 'flags must be empty if regex is passed pre-compiled'
            return regex
    return SimpleLazyObject(_compile)


@deconstructible
class JSONValidator:
    """
    Validates that the input is JSON compliant.
    """
    def __call__(self, value):
        value = value.strip()
        try:
            json.loads(s=value)
        except ValueError:
            raise ValidationError(
                code='invalid',
                message=_(message='Enter a valid JSON value.')
            )

    def __eq__(self, other):
        return (
            isinstance(other, JSONValidator)
        )

    def __ne__(self, other):
        return not (self == other)


@deconstructible
class YAMLValidator:
    """
    Validates that the input is YAML compliant.
    """
    def __call__(self, value):
        value = value.strip()
        try:
            yaml.load(stream=value, Loader=SafeLoader)
        except yaml.error.YAMLError:
            raise ValidationError(
                code='invalid',
                message=_(message='Enter a valid YAML value.')
            )

    def __eq__(self, other):
        return (
            isinstance(other, YAMLValidator)
        )

    def __ne__(self, other):
        return not (self == other)


internal_name_re = _lazy_re_compile(regex=r'^[a-zA-Z0-9_]+\Z')
validate_internal_name = RegexValidator(
    internal_name_re, _(
        "Enter a valid 'internal name' consisting of letters, numbers, and "
        "underscores."
    ), 'invalid'
)



from django.apps import apps


def handler_initialize_new_document_type_file_metadata_driver_configuration(
    sender, instance, **kwargs
):
    DocumentTypeDriverConfiguration = apps.get_model(
        app_label='file_metadata',
        model_name='DocumentTypeDriverConfiguration'
    )
    StoredDriver = apps.get_model(
        app_label='file_metadata', model_name='StoredDriver'
    )

    if kwargs['created']:
        for stored_driver in StoredDriver.objects.all():
            driver_class = stored_driver.driver_class
            enabled = driver_class.get_enabled_value()

            DocumentTypeDriverConfiguration.objects.create(
                document_type=instance, enabled=enabled,
                stored_driver=stored_driver
            )


def handler_post_document_file_upload(sender, instance, **kwargs):
    instance.submit_for_file_metadata_processing()



from django.urls import re_path

from .api_views.saved_resultset_api_views import (
    APISavedResultsetCreateView, APISavedResultsetDetailView,
    APISavedResultsetListView, APISavedResultsetResultListView
)
from .api_views.search_api_views import (
    APISearchModelDetailView, APISearchModelList, APISearchView
)
from .views.saved_resultset_views import (
    SavedResultsetDeleteView, SavedResultsetListView,
    SavedResultsetResultListView
)
from .views.search_views import (
    SearchAdvancedView, SearchAgainView, SearchBackendReindexView,
    SearchResultView, SearchSimpleView
)
from .views.tool_views import (
    SearchModelListView, SearchModelSearchFieldListView
)

urlpatterns_search = [
    re_path(
        route=r'^again/(?P<search_model_pk>[\.\w]+)/$', name='search_again',
        view=SearchAgainView.as_view()
    ),
    re_path(
        route=r'^advanced/(?P<search_model_pk>[\.\w]+)/$',
        name='search_advanced', view=SearchAdvancedView.as_view()
    ),
    re_path(
        route=r'^advanced/$', name='search_advanced',
        view=SearchAdvancedView.as_view()
    ),
    re_path(
        route=r'^results/$', name='search_results',
        view=SearchResultView.as_view()
    ),
    re_path(
        route=r'^results/(?P<search_model_pk>[\.\w]+)/$',
        name='search_results', view=SearchResultView.as_view()
    ),
    re_path(
        route=r'^simple/(?P<search_model_pk>[\.\w]+)/$',
        name='search_simple', view=SearchSimpleView.as_view()
    )
]

urlpatterns_saved_resultset = [
    re_path(
        route=r'^saved_resultsets/$',
        name='saved_resultset_list', view=SavedResultsetListView.as_view()
    ),
    re_path(
        route=r'^saved_resultsets/(?P<saved_resultset_id>\d+)/delete/$',
        name='saved_resultset_delete_single',
        view=SavedResultsetDeleteView.as_view()
    ),
    re_path(
        route=r'^saved_resultsets/(?P<saved_resultset_id>\d+)/results/$',
        name='saved_resultset_result_list',
        view=SavedResultsetResultListView.as_view()
    )
]

urlpatterns_tools = [
    re_path(
        route=r'^backend/reindex/$', name='search_backend_reindex',
        view=SearchBackendReindexView.as_view()
    ),
    re_path(
        route=r'^search_models/$', name='search_model_list',
        view=SearchModelListView.as_view()
    ),
    re_path(
        route=r'^search_models/(?P<search_model_name>[\w.]+)/search_fields/$',
        name='search_model_detail',
        view=SearchModelSearchFieldListView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_saved_resultset)
urlpatterns.extend(urlpatterns_search)
urlpatterns.extend(urlpatterns_tools)

api_urls_saved_resultset = [
    re_path(
        route=r'^saved_resultsets/(?P<saved_resultset_id>[0-9]+)/$',
        view=APISavedResultsetDetailView.as_view(),
        name='saved_resultset-detail'
    ),
    re_path(
        route=r'^saved_resultsets/(?P<saved_resultset_id>[0-9]+)/results/$',
        view=APISavedResultsetResultListView.as_view(),
        name='saved_resultset-result-list'
    ),
    re_path(
        route=r'^saved_resultsets/$', name='saved_resultset-list',
        view=APISavedResultsetListView.as_view()
    ),
    re_path(
        route=r'^saved_resultsets/(?P<search_model_pk>[\.\w]+)/$',
        name='saved_resultset-create',
        view=APISavedResultsetCreateView.as_view()
    )
]

api_urls_search = [
    re_path(
        route=r'^search/(?P<search_model_pk>[\.\w]+)/$', name='search-view',
        view=APISearchView.as_view()
    ),
    re_path(
        route=r'^search/advanced/(?P<search_model_pk>[\.\w]+)/$',
        name='advanced-search-view', view=APISearchView.as_view()
    ),
    re_path(
        route=r'^search_models/$', name='searchmodel-list',
        view=APISearchModelList.as_view()
    ),
    re_path(
        route=r'^search_models/(?P<search_model_pk>[\.\w]+)/$',
        name='searchmodel-detail', view=APISearchModelDetailView.as_view()
    )
]

api_urls = []
api_urls.extend(api_urls_saved_resultset)
api_urls.extend(api_urls_search)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0011_populate_internal_name')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='filemetadataentry', options={
                'ordering': ('internal_name', 'value'),
                'verbose_name': 'File metadata entry',
                'verbose_name_plural': 'File metadata entries'
            }
        ),
        migrations.AlterUniqueTogether(
            name='filemetadataentry', unique_together={
                ('document_file_driver_entry', 'internal_name')
            }
        )
    ]



from mayan.apps.smart_settings.setting_namespaces import (
    SettingNamespaceMigration
)
from mayan.apps.smart_settings.utils import smart_yaml_load


class ConvertSettingMigration(SettingNamespaceMigration):
    """
    From version 0001 to 0002 backend arguments are no longer quoted
    but YAML valid too. Changed in version 3.3.
    """
    def converter_graphics_backend_arguments_0001(self, value):
        return smart_yaml_load(value=value)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import DEFAULT_EMAIL, DEFAULT_PASSWORD, DEFAULT_USERNAME

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Auto administrator'), name='autoadmin'
)

setting_email = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL, global_name='AUTOADMIN_EMAIL', help_text=_(
        message='Sets the email of the automatically created super user account.'
    )
)
setting_password = setting_namespace.do_setting_add(
    default=DEFAULT_PASSWORD, global_name='AUTOADMIN_PASSWORD', help_text=_(
        message='The password of the automatically created super user account. '
        'If it is equal to None, the password is randomly generated.'
    )
)
setting_username = setting_namespace.do_setting_add(
    default=DEFAULT_USERNAME, global_name='AUTOADMIN_USERNAME', help_text=_(
        message='The username of the automatically created super user account.'
    )
)



from elasticsearch.dsl import Q, field

from ...search_query_types import (
    BackendQueryType, QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan, QueryTypeLessThanOrEqual,
    QueryTypePartial, QueryTypeRange, QueryTypeRangeExclusive,
    QueryTypeRegularExpression
)

from .backend import ElasticsearchSearchBackend


class BackendQueryTypeExact(BackendQueryType):
    query_type = QueryTypeExact

    def do_resolve(self):
        if self.value is not None:
            if self.is_quoted_value:
                template = '"{}"'
            else:
                template = '{}'

            if not self.value:
                # Empty values cannot be quoted.
                template = '{}'

                if self.is_quoted_value:
                    if self.get_search_backend_field_type() == field.Text:
                        return Q(
                            'bool', must_not=(
                                Q(
                                    'wildcard',
                                    **{self.search_field.field_name: '*'}
                                )
                            )
                        )

            if self.is_quoted_value:
                return Q(
                    name_or_query='match_phrase', _expand__to_dot=False,
                    **{
                        self.search_field.field_name: template.format(self.value)
                    }
                )
            else:
                return Q(
                    name_or_query='match', _expand__to_dot=False,
                    **{
                        self.search_field.field_name: template.format(self.value)
                    }
                )


class BackendQueryFuzzy(BackendQueryType):
    query_type = QueryTypeFuzzy

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='fuzzy', _expand__to_dot=False,
                **{self.search_field.field_name: self.value}
            )


class BackendQueryTypeGreaterThan(BackendQueryType):
    query_type = QueryTypeGreaterThan

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'gt': self.value}
                }
            )


class BackendQueryTypeGreaterThanOrEqual(BackendQueryType):
    query_type = QueryTypeGreaterThanOrEqual

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'gte': self.value}
                }
            )


class BackendQueryTypeLessThan(BackendQueryType):
    query_type = QueryTypeLessThan

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'lt': self.value}
                }
            )


class BackendQueryTypeLessThanOrEqual(BackendQueryType):
    query_type = QueryTypeLessThanOrEqual

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'lte': self.value}
                }
            )


class BackendQueryTypePartial(BackendQueryType):
    query_type = QueryTypePartial

    def do_resolve(self):
        if self.value is not None:
            if self.get_search_backend_field_type() != field.Date:
                if self.is_quoted_value:
                    return Q(
                        name_or_query='match_phrase', _expand__to_dot=False,
                        **{
                            self.search_field.field_name: '{}'.format(self.value)
                        }
                    )
                else:
                    if self.get_search_backend_field_type() != field.Integer:
                        return Q(
                            name_or_query='wildcard', _expand__to_dot=False,
                            **{
                                self.search_field.field_name: '*{}*'.format(self.value)
                            }
                        )


class BackendQueryTypeRange(BackendQueryType):
    query_type = QueryTypeRange

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {
                        'gte': self.value[0], 'lte': self.value[1]
                    }
                }
            )


class BackendQueryTypeRangeExclusive(BackendQueryType):
    query_type = QueryTypeRangeExclusive

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {
                        'gt': self.value[0], 'lt': self.value[1]
                    }
                }
            )


class BackendQueryTypeRegularExpression(BackendQueryType):
    query_type = QueryTypeRegularExpression

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='regexp', _expand__to_dot=False,
                **{self.search_field.field_name: self.value}
            )


BackendQueryType.register(
    klass=BackendQueryTypeExact, search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryFuzzy, search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThan, search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThanOrEqual,
    search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThan, search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThanOrEqual,
    search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypePartial, search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRange,
    search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRangeExclusive,
    search_backend=ElasticsearchSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRegularExpression,
    search_backend=ElasticsearchSearchBackend
)



from django.urls import resolve as django_resolve
from django.urls.base import get_script_prefix

from .literals import URL_QUERY_POSITIVE_VALUES


def convert_to_id_list(items):
    return ','.join(
        map(str, items)
    )


def get_request_data(request):
    request_get_data = request.GET.dict()
    request_post_data = request.POST.dict()

    query_dict = request_get_data.copy()
    query_dict.update(request_post_data)

    return query_dict


def is_url_query_positive(value):
    if value is not None:
        return value.lower() in URL_QUERY_POSITIVE_VALUES


def request_is_ajax(request):
    return request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest'


def resolve(path, urlconf=None):
    path = '/{}'.format(
        path.replace(
            get_script_prefix(), '', 1
        )
    )
    return django_resolve(path=path, urlconf=urlconf)



import hashlib

from django.core import serializers
from django.utils.translation import gettext_lazy as _

from ..literals import GRAPHVIZ_SYMBOL_CONDITIONAL, GRAPHVIZ_SYMBOL_TIME


class WorkflowStateEscalationBusinessLogicMixin:
    def do_diagram_generate(self, diagram):
        if self.enabled:
            escalation_label = '{} {}'.format(
                GRAPHVIZ_SYMBOL_TIME, self.get_time_display()
            )
            if self.has_condition():
                escalation_label = '{} {}'.format(
                    escalation_label, GRAPHVIZ_SYMBOL_CONDITIONAL
                )

            edge_kwargs = {
                'head_name': self.transition.destination_state.get_graph_id(),
                'label': '''<
                    <table border="3" cellborder="0" color="white">
                        <tr>
                            <td bgcolor="white">{}</td>
                        </tr>
                    </table>
                >'''.format(escalation_label,),
                'style': 'dashed',
                'tail_name': self.state.get_graph_id()
            }
            diagram.edge(**edge_kwargs)

    def get_comment(self):
        return self.comment or _(message='Workflow escalation.')

    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()

    def get_time_display(self):
        return '{} {}'.format(self.amount, self.unit)



from mayan.apps.smart_settings.setting_namespaces import (
    SettingNamespaceMigration
)


class SettingMigrationMIMEType(SettingNamespaceMigration):
    def mime_type_backend_0001(self, value):
        """
        The Python Magic backed was removed in version 4.9, this migration
        switches the backend to the file based one.
        """
        return 'mayan.apps.mime_types.backends.file_command.MIMETypeBackendFileCommand'



from django.apps import apps
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.sources.classes import DocumentCreateWizardStep
from mayan.apps.sources.wizard_steps import (
    DocumentCreateWizardStepDocumentType
)
from mayan.apps.views.http import URL

from .forms import TagMultipleSelectionForm
from .models import Tag
from .permissions import permission_tag_attach


class DocumentCreateWizardStepTags(DocumentCreateWizardStep):
    form_class = TagMultipleSelectionForm
    label = _(message='Select tags')
    name = 'tag_selection'
    number = 2

    @staticmethod
    def get_tags_queryset_for_document(document, user=None):
        queryset_documents = Document.valid.all()

        if user:
            queryset_documents = AccessControlList.objects.restrict_queryset(
                permission=permission_tag_attach,
                queryset=queryset_documents, user=user
            )

        try:
            queryset_documents.get(pk=document.pk)
        except Document.DoesNotExist:
            queryset_tags = Tag.objects.none()
        else:
            queryset_tags = Tag.objects.all()

            if user:
                queryset_tags = AccessControlList.objects.restrict_queryset(
                    permission=permission_tag_attach, queryset=queryset_tags,
                    user=user
                )

        return queryset_tags

    @staticmethod
    def get_tags_queryset_for_document_type(document_type, user):
        queryset_document_types = DocumentType.objects.all()

        queryset_document_types_restricted = AccessControlList.objects.restrict_queryset(
            permission=permission_tag_attach,
            queryset=queryset_document_types, user=user
        )

        try:
            queryset_document_types_restricted.get(pk=document_type.pk)
        except DocumentType.DoesNotExist:
            queryset_tags = Tag.objects.none()
        else:
            queryset_tags = Tag.objects.all()

        return queryset_tags

    @classmethod
    def condition(cls, wizard):
        Tag = apps.get_model(app_label='tags', model_name='Tag')

        return Tag.objects.exists()

    @classmethod
    def get_form_kwargs(self, wizard):
        step_data = wizard.get_cleaned_data_for_step(
            step=DocumentCreateWizardStepDocumentType.name
        )
        user = wizard.request.user

        kwargs = {
            'help_text': _(message='Tags to be attached.'),
            'permission': permission_tag_attach,
            'queryset': Tag.objects.none(),
            'user': user
        }

        if step_data:
            document_type = step_data['document_type']

            queryset_tags = DocumentCreateWizardStepTags.get_tags_queryset_for_document_type(
                document_type=document_type, user=user
            )

            kwargs['queryset'] = queryset_tags

        return kwargs

    @classmethod
    def done(cls, wizard):
        result = {}
        cleaned_data = wizard.get_cleaned_data_for_step(cls.name)
        if cleaned_data:
            result['tags'] = [
                str(tag.pk) for tag in cleaned_data['tags']
            ]

        return result

    @classmethod
    def step_post_upload_process(
        cls, document, query_string, source_id, user_id
    ):
        User = get_user_model()

        try:
            user = User.objects.get(pk=user_id)
        except User.DoesNotExist:
            user = None

        queryset_tags = DocumentCreateWizardStepTags.get_tags_queryset_for_document(
            document=document, user=user
        )

        tag_id_list = URL(query_string=query_string).args.getlist('tags')

        for tag in queryset_tags.filter(pk__in=tag_id_list):
            if user:
                tag.attach_to(document=document, user=user)
            else:
                tag._attach_to(document=document)


DocumentCreateWizardStep.register(step=DocumentCreateWizardStepTags)



from rest_framework import status

from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentFileAPIViewMixin
)
from mayan.apps.rest_api import generics

from ..permissions import (
    permission_file_metadata_submit, permission_file_metadata_view
)
from ..serializers.document_file_serializers import (
    DocumentFileMetadataDriverEntrySerializer,
    DocumentFileMetadataEntrySerializer
)

from .api_view_mixins import ParentObjectDocumentFileDriverAPIViewMixin


class APIDocumentFileMetadataDriverListView(
    ParentObjectDocumentFileAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of selected document's file metadata drivers.
    """
    serializer_class = DocumentFileMetadataDriverEntrySerializer

    def get_source_queryset(self):
        document_file = self.get_document_file(
            permission=permission_file_metadata_view
        )
        return document_file.file_metadata_drivers.all()


class APIDocumentFileMetadataDriverDetailView(
    ParentObjectDocumentFileAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns the details of the document file metadata driver.
    """
    lookup_url_kwarg = 'driver_id'
    serializer_class = DocumentFileMetadataDriverEntrySerializer

    def get_source_queryset(self):
        document_file = self.get_document_file(
            permission=permission_file_metadata_view
        )
        return document_file.file_metadata_drivers.all()


class APIDocumentFileMetadataEntryListView(
    ParentObjectDocumentFileDriverAPIViewMixin, generics.ListAPIView
):
    """
    get: Return the list of file metadata entries for the selected driver.
    """
    serializer_class = DocumentFileMetadataEntrySerializer

    def get_source_queryset(self):
        document_file_metadata_driver = self.get_document_file_metadata_driver(
            permission=permission_file_metadata_view
        )
        return document_file_metadata_driver.entries.all()


class APIDocumentFileMetadataEntryDetailView(
    ParentObjectDocumentFileDriverAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Return the detail of selected document file metadata entry.
    """
    lookup_url_kwarg = 'entry_id'
    serializer_class = DocumentFileMetadataEntrySerializer

    def get_source_queryset(self):
        document_file_metadata_driver = self.get_document_file_metadata_driver(
            permission=permission_file_metadata_view
        )
        return document_file_metadata_driver.entries.all()


class APIDocumentFileMetadataSubmitView(
    ParentObjectDocumentFileDriverAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Submit a document file for file metadata processing.
    """
    action_response_status = status.HTTP_202_ACCEPTED
    lookup_url_kwarg = 'document_file_id'
    mayan_object_permission_map = {'POST': permission_file_metadata_submit}

    def get_source_queryset(self):
        return self.get_document_file_queryset()

    def object_action(self, obj, request, serializer):
        obj.submit_for_file_metadata_processing(user=self.request.user)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_SIGNATURES_BACKEND, DEFAULT_DEFAULT_GPG_PATH,
    DEFAULT_SIGNATURES_KEYSERVER
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Signatures'), name='django_gpg'
)

setting_gpg_backend = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURES_BACKEND,
    global_name='SIGNATURES_BACKEND',
    help_text=_(
        message='Full path to the backend to be used to handle keys and signatures.'
    )
)
setting_gpg_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_DEFAULT_GPG_PATH,
    global_name='SIGNATURES_BACKEND_ARGUMENTS',
)
setting_keyserver = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURES_KEYSERVER, global_name='SIGNATURES_KEYSERVER',
    help_text=_(message='Keyserver used to query for keys.')
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='OCR'), name='ocr'
)

permission_document_version_ocr = namespace.add_permission(
    label=_(message='Submit documents for OCR'), name='ocr_document'
)
permission_document_version_ocr_content_edit = namespace.add_permission(
    label=_(message='Edit the transcribed text from document'),
    name='ocr_content_edit'
)
permission_document_version_ocr_content_view = namespace.add_permission(
    label=_(message='View the transcribed text from document'),
    name='ocr_content_view'
)
permission_document_type_ocr_setup = namespace.add_permission(
    label=_(message='Change document type OCR settings'),
    name='ocr_document_type_setup'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_EVENTS_DISABLE_ASYNCHRONOUS_MODE, DEFAULT_EVENTS_PRUNE_BACKEND,
    DEFAULT_EVENTS_PRUNE_BACKEND_ARGUMENTS,
    DEFAULT_EVENTS_PRUNE_TASK_INTERVAL
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Events'), name='events'
)

setting_disable_asynchronous_mode = setting_namespace.do_setting_add(
    default=DEFAULT_EVENTS_DISABLE_ASYNCHRONOUS_MODE,
    global_name='EVENTS_DISABLE_ASYNCHRONOUS_MODE',
    help_text=_(
        message='Disables asynchronous events mode. All events will be '
        'committed in the same process that triggers them. This was the '
        'behavior prior to version 4.5.'
    )
)
setting_event_prune_backend = setting_namespace.do_setting_add(
    default=DEFAULT_EVENTS_PRUNE_BACKEND, global_name='EVENTS_PRUNE_BACKEND',
    help_text=_(
        'Path to the event pruning subclass that will be called '
        'periodically to clear the event log.'
    )
)
setting_event_prune_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_EVENTS_PRUNE_BACKEND_ARGUMENTS,
    global_name='EVENTS_PRUNE_BACKEND_ARGUMENTS',
    help_text=_('Arguments to pass to `EVENTS_PRUNE_BACKEND`.')
)
setting_event_prune_task_interval = setting_namespace.do_setting_add(
    default=DEFAULT_EVENTS_PRUNE_TASK_INTERVAL,
    global_name='EVENTS_PRUNE_TASK_INTERVAL',
    help_text=_(
        'Time interval in seconds, at which the event prune task will '
        'execute.'
    )
)



from django.urls import include, re_path

from mayan.apps.backends.classes import BaseBackend
from mayan.apps.common.menus import menu_tools

from .settings import (
    setting_client_backend_arguments, setting_client_backend_enabled
)


class ClientBackend(BaseBackend):
    _loader_module_name = 'client_backends'

    @classmethod
    def get_backend_instance(cls, name):
        backend_class = cls.get(name=name)
        kwargs = setting_client_backend_arguments.value.get(
            name, {}
        )
        return backend_class(**kwargs)

    @classmethod
    def post_load_modules(cls):
        cls.register_url_patterns()
        cls.register_links()
        cls.launch_backends()

    @classmethod
    def launch_backends(cls):
        for backend_name in setting_client_backend_enabled.value:
            cls.get_backend_instance(name=backend_name).launch()

    @classmethod
    def register_links(cls):
        for backend_name in setting_client_backend_enabled.value:
            backend_instance = cls.get_backend_instance(name=backend_name)
            menu_tools.bind_links(
                links=backend_instance.get_links()
            )

    @classmethod
    def register_url_patterns(cls):
        # Hidden import.
        from .urls import urlpatterns

        for backend_name in setting_client_backend_enabled.value:
            backend_instance = cls.get_backend_instance(name=backend_name)

            top_url = '{}/'.format(
                getattr(
                    backend_instance, '_url_namespace',
                    backend_instance.__class__.__name__
                )
            )

            urlpatterns += (
                re_path(
                    route=r'^{}'.format(top_url), view=include(
                        backend_instance.get_url_patterns()
                    )
                ),
            )

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def get_links(self):
        return ()



from django.utils.translation import gettext_lazy as _

from .classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Permissions'), name='permissions'
)

permission_role_create = namespace.add_permission(
    label=_(message='Create roles'), name='role_create'
)
permission_role_delete = namespace.add_permission(
    label=_(message='Delete roles'), name='role_delete'
)
permission_role_edit = namespace.add_permission(
    label=_(message='Edit roles'), name='role_edit'
)
permission_role_view = namespace.add_permission(
    label=_(message='View roles'), name='role_view'
)



from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.rest_api import serializers

from ..classes import EventType
from ..models import StoredEventType


class EventTypeNamespaceSerializer(serializers.Serializer):
    event_types_url = serializers.HyperlinkedIdentityField(
        lookup_field='name',
        view_name='rest_api:event-type-namespace-event-type-list'
    )
    label = serializers.CharField(
        label=_(message='Label')
    )
    name = serializers.CharField(
        label=_(message='Name')
    )
    url = serializers.SerializerMethodField(
        label=_(message='URL')
    )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:event-type-namespace-detail', kwargs={
                'name': instance.name
            }, request=self.context['request'], format=self.context['format']
        )


class EventTypeSerializer(serializers.Serializer):
    event_type_namespace_url = serializers.SerializerMethodField(
        label=_(message='Event type namespace URL')
    )
    id = serializers.CharField(
        label=_(message='ID')
    )
    label = serializers.CharField(
        label=_(message='Label')
    )
    name = serializers.CharField(
        label=_(message='Name')
    )

    def get_event_type_namespace_url(self, instance):
        return reverse(
            viewname='rest_api:event-type-namespace-detail', kwargs={
                'name': instance.namespace.name
            }, request=self.context['request'], format=self.context['format']
        )

    def to_representation(self, instance):
        if isinstance(instance, EventType):
            return super().to_representation(instance=instance)
        elif isinstance(instance, StoredEventType):
            return super().to_representation(instance=instance.event_type)
        elif isinstance(instance, str):
            return super().to_representation(
                instance=EventType.get(id=instance)
            )



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0035_workflowstate_final')
    ]

    operations = [
        migrations.AddField(
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='workflow_instances',
                to='document_states.workflowstate',
                verbose_name='Active state'
            ), model_name='workflowinstance', name='state_active'
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0008_auto_20180315_0029')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='objecteventsubscription',
            options={
                'ordering': ('pk',),
                'verbose_name': 'Object event subscription',
                'verbose_name_plural': 'Object event subscriptions'
            }
        )
    ]



from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .models import ResolvedWebLink, WebLink


class WebLinkDocumentTypeAddSerializer(serializers.Serializer):
    document_type = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the document type to add to the web link.'
        ), label=_(message='Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )


class WebLinkDocumentTypeRemoveSerializer(serializers.Serializer):
    document_type = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the document type to remove from the web '
            'link.'
        ), label=_(message='Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )


class WebLinkSerializer(serializers.HyperlinkedModelSerializer):
    document_types_add_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document types add URL'),
        lookup_url_kwarg='web_link_id',
        view_name='rest_api:web_link-document_type-add'
    )
    document_types_remove_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document types remove URL'),
        lookup_url_kwarg='web_link_id',
        view_name='rest_api:web_link-document_type-remove'
    )
    document_types_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document types URL'), lookup_url_kwarg='web_link_id',
        view_name='rest_api:web_link-document_type-list'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'web_link_id',
                'view_name': 'rest_api:web_link-detail'
            }
        }
        fields = (
            'document_types_add_url', 'document_types_remove_url',
            'document_types_url', 'enabled', 'id', 'label', 'template',
            'url'
        )
        model = WebLink


class ResolvedWebLinkSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.SerializerMethodField(
        label=_(message='URL')
    )
    navigation_url = serializers.SerializerMethodField(
        label=_(message='Navigation URL')
    )

    class Meta:
        fields = ('id', 'navigation_url', 'url')
        model = ResolvedWebLink

    def get_url(self, instance):
        return reverse(
            format=self.context['format'], kwargs={
                'document_id': self.context['external_object'].pk,
                'resolved_web_link_id': instance.pk
            }, request=self.context['request'],
            viewname='rest_api:resolved_web_link-detail'
        )

    def get_navigation_url(self, instance):
        return reverse(
            format=self.context['format'], kwargs={
                'document_id': self.context['external_object'].pk,
                'resolved_web_link_id': instance.pk
            }, request=self.context['request'],
            viewname='rest_api:resolved_web_link-navigate'
        )



from mayan.apps.documents.models.document_models import Document
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..permissions import (
    permission_document_metadata_add, permission_document_metadata_edit,
    permission_document_metadata_remove, permission_document_metadata_view
)
from ..serializers import DocumentMetadataSerializer


class APIDocumentMetadataListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of selected document's metadata types and values.
    post: Add an existing metadata type and value to the selected document.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permission_map = {
        'GET': permission_document_metadata_view,
        'POST': permission_document_metadata_add
    }
    mayan_object_permission_map = {'GET': permission_document_metadata_view}
    serializer_class = DocumentMetadataSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'document': self.get_external_object()
        }

    def get_source_queryset(self):
        return self.get_external_object().metadata.all()

    def perform_create(self, serializer):
        if 'metadata_type_id' in serializer.validated_data:
            serializer.validated_data['metadata_type'] = serializer.validated_data['metadata_type_id']

        return super().perform_create(serializer=serializer)


class APIDocumentMetadataView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Remove this metadata entry from the selected document.
    get: Return the details of the selected document metadata type and value.
    patch: Edit the selected document metadata type and value.
    put: Edit the selected document metadata type and value.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    lookup_url_kwarg = 'metadata_id'
    mayan_external_object_permission_map = {
        'DELETE': permission_document_metadata_remove,
        'GET': permission_document_metadata_view,
        'PATCH': permission_document_metadata_edit,
        'PUT': permission_document_metadata_edit
    }
    mayan_object_permission_map = {
        'DELETE': permission_document_metadata_remove,
        'GET': permission_document_metadata_view,
        'PATCH': permission_document_metadata_edit,
        'PUT': permission_document_metadata_edit
    }
    serializer_class = DocumentMetadataSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.get_external_object().metadata.all()



import base64

from django.template import Library, Node, TemplateSyntaxError
from django.utils.html import strip_spaces_between_tags
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.utils import flatten_map, flatten_object

from ..decorators import templating_dangerous_tag

register = Library()


# Filters


@register.filter(name='dict_get')
def filter_dict_get(dictionary, key):
    """
    Return the value for the given key or '' if not found.
    Deprecated in favor or `dictionary_get`.
    """
    return dictionary.get(key, '')


@register.filter(name='dictionary_flatten')
def filter_dictionary_flatten(dictionary):
    """
    Return a flat version of a nested dictionary.
    """
    result = {}

    flatten_map(dictionary=dictionary, result=result, separator='__')

    return result


@register.filter(name='dictionary_get')
def filter_dictionary_get(dictionary, key):
    """
    Return the value for the given key or '' if not found.
    """
    return dictionary.get(key, '')


@register.filter(name='object_flatten')
def filter_object_flatten(value):
    """
    Return a flat version of a nested object of multiple types.
    """

    result = dict(
        flatten_object(obj=value, separator='__')
    )

    return result


@register.filter(name='split')
def filter_split(obj, separator):
    """
    Return a list of the words in the string, using sep as the delimiter
    string.
    """
    return obj.split(separator)


@register.filter(name='to_base64')
def filter_to_base64(value, altchars=None):
    """
    Convert a value to base64 encoding. Accepts optional `altchars` argument.
    """
    if altchars:
        altchars = bytes(encoding='utf-8', source=altchars)
    return base64.b64encode(s=value, altchars=altchars).decode('utf-8')


# Tags


class SpacelessPlusNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        content = self.nodelist.render(context).strip()
        result = []
        for line in content.split('\n'):
            if line.strip() != '':
                result.append(line)

        return strip_spaces_between_tags(
            value='\n'.join(result)
        )


@register.simple_tag(name='method')
@templating_dangerous_tag(
    reason=_(
        message='Tag can bypass access controls and create permanent changes.'
    )
)
def tag_method(obj, method, *args, **kwargs):
    """
    Call an object method. {% method object method **kwargs %}
    """
    try:
        return getattr(obj, method)(*args, **kwargs)
    except Exception as exception:
        raise TemplateSyntaxError(
            'Error calling object method; {}'.format(exception)
        )


@register.simple_tag(name='range')
def tag_range(*args):
    """
    Return an object that produces a sequence of integers
    """
    return range(*args)


@register.simple_tag(name='set')
def tag_set(value):
    """
    Set a context variable to a specific value.
    """
    return value


@register.tag(name='spaceless_plus')
def tag_spaceless_plus(parser, token):
    """
    Removes empty lines between the tag nodes.
    """
    nodelist = parser.parse(
        ('endspaceless_plus',)
    )
    parser.delete_first_token()
    return SpacelessPlusNode(nodelist=nodelist)



from datetime import timedelta

from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_e

from .literals import TASK_SAVED_RESULTSET_EXPIRED_DELETE_INTERVAL

queue_search = CeleryQueue(
    label=_(message='Search'), name='search', worker=worker_e
)
queue_search_slow = CeleryQueue(
    label=_(message='Search slow'), name='search_slow', worker=worker_e
)

queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_deindex_instance',
    label=_(message='Remove a model instance from the search engine.'),
    name='task_deindex_instance'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_index_instance',
    label=_(message='Index a model instance to the search engine.'),
    name='task_index_instance'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_index_instances',
    label=_(
        message='Index all instances of a search model to the search engine.'
    ), name='task_index_instances'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_index_related_instance_m2m',
    label=_(
        message='Index all related instances of a search model after a many '
        'to many event.'
    ), name='task_index_related_instance_m2m'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_saved_resultset_expired_delete',
    label=_(message='Delete expired saved resultsets'),
    name='task_saved_resultset_expired_delete',
    schedule=timedelta(seconds=TASK_SAVED_RESULTSET_EXPIRED_DELETE_INTERVAL)
)

queue_search_slow.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_reindex_backend',
    label=_(
        message='Reset the search backend indices and index all instances '
        'again.'
    ), name='task_reindex_backend'
)



from actstream.models import Action, any_stream

from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import (
    ExternalContentTypeObjectAPIViewMixin
)

from ..permissions import permission_events_view
from ..serializers.event_serializers import EventSerializer


class APIObjectEventListView(
    ExternalContentTypeObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of events for the specified object.
    """
    mayan_external_object_permission_map = {'GET': permission_events_view}
    serializer_class = EventSerializer

    def get_source_queryset(self):
        return any_stream(
            obj=self.get_external_object()
        )


class APIEventListView(generics.ListAPIView):
    """
    get: Returns a list of all the available events.
    """
    mayan_view_permission_map = {'GET': permission_events_view}
    serializer_class = EventSerializer
    source_queryset = Action.objects.all()

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }



from mayan.apps.common.serialization import yaml_dump
from mayan.apps.forms import form_widgets, forms

from .models import DocumentTypeDriverConfiguration


class FormDocumentTypeFileMetadataDriverConfiguration(forms.ModelForm):
    class Meta:
        fields = ('enabled', 'arguments')
        model = DocumentTypeDriverConfiguration

    def __init__(self, driver_configuration=None, *args, **kwargs):
        self._driver_configuration = driver_configuration

        super().__init__(*args, **kwargs)

        if self.instance:
            arguments = self.instance.get_arguments()

            for key, value in arguments.items():
                self.initial[key] = value

        self.fields['arguments'].widget = form_widgets.HiddenInput()

    def clean(self):
        # Otherwise grab the values from the dynamic form and create
        # the argument JSON object.
        result = {}

        argument_name_list = self.instance.stored_driver.driver_class.get_argument_name_list()

        for argument_name in argument_name_list:
            if self.cleaned_data[argument_name] is not None:
                result[argument_name] = self.cleaned_data[argument_name]

        self.cleaned_data['arguments'] = yaml_dump(data=result)



from django.urls import re_path

from .api_views import (
    APIResolvedWebLinkListView, APIResolvedWebLinkNavigateView,
    APIResolvedWebLinkView, APIWebLinkDocumentTypeAddView,
    APIWebLinkDocumentTypeListView, APIWebLinkDocumentTypeRemoveView,
    APIWebLinkListView, APIWebLinkView
)
from .views import (
    DocumentTypeWebLinksView, DocumentWebLinkListView, ResolvedWebLinkView,
    WebLinkCreateView, WebLinkDeleteView, WebLinkDocumentTypesView,
    WebLinkEditView, WebLinkListView
)

urlpatterns_document_types = [
    re_path(
        route=r'^document_types/(?P<document_type_id>\d+)/web_links/$',
        name='document_type_web_links',
        view=DocumentTypeWebLinksView.as_view()
    )
]

urlpatterns_documents = [
    re_path(
        route=r'^documents/(?P<document_id>\d+)/web_links/$',
        name='document_web_link_list', view=DocumentWebLinkListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/web_links/(?P<web_link_id>\d+)/$',
        name='web_link_instance_view', view=ResolvedWebLinkView.as_view()
    )
]

urlpatterns_web_links = [
    re_path(
        route=r'^web_links/$', name='web_link_list',
        view=WebLinkListView.as_view()
    ),
    re_path(
        route=r'^web_links/create/$', name='web_link_create',
        view=WebLinkCreateView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>\d+)/delete/$',
        name='web_link_delete', view=WebLinkDeleteView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>\d+)/document_types/$',
        name='web_link_document_types',
        view=WebLinkDocumentTypesView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>\d+)/edit/$', name='web_link_edit',
        view=WebLinkEditView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_document_types)
urlpatterns.extend(urlpatterns_documents)
urlpatterns.extend(urlpatterns_web_links)

api_urls = [
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/resolved_web_links/$',
        name='resolved_web_link-list',
        view=APIResolvedWebLinkListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/resolved_web_links/(?P<resolved_web_link_id>[0-9]+)/$',
        name='resolved_web_link-detail',
        view=APIResolvedWebLinkView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/resolved_web_links/(?P<resolved_web_link_id>[0-9]+)/navigate/$',
        name='resolved_web_link-navigate',
        view=APIResolvedWebLinkNavigateView.as_view()
    ),
    re_path(
        route=r'^web_links/$', name='web_link-list',
        view=APIWebLinkListView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>[0-9]+)/$',
        name='web_link-detail', view=APIWebLinkView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>[0-9]+)/document_types/$',
        name='web_link-document_type-list',
        view=APIWebLinkDocumentTypeListView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>[0-9]+)/document_types/add/$',
        name='web_link-document_type-add',
        view=APIWebLinkDocumentTypeAddView.as_view()
    ),
    re_path(
        route=r'^web_links/(?P<web_link_id>[0-9]+)/document_types/remove/$',
        name='web_link-document_type-remove',
        view=APIWebLinkDocumentTypeRemoveView.as_view()
    )
]



from furl import furl

from django.apps import apps
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import classonlymethod
from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import forms, wizards
from mayan.apps.views.view_mixins import ViewIconMixin

from .classes import DocumentCreateWizardStep
from .exceptions import SourceActionExceptionUnknown
from .icons import (
    icon_document_upload_wizard, icon_wizard_step_first,
    icon_wizard_step_next, icon_wizard_step_previous
)


class DocumentCreateWizard(ViewIconMixin, wizards.SessionWizardView):
    template_name = 'appearance/wizard.html'
    view_icon = icon_document_upload_wizard

    @classonlymethod
    def as_view(cls, *args, **kwargs):
        # SessionWizardView needs at least one form in order to be
        # initialized as a view. Declare one empty form and then change the
        # form list in the .dispatch() method.
        class EmptyForm(forms.Form):
            """Empty form."""

        cls.form_list = [EmptyForm]
        return super().as_view(*args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        Source = apps.get_model(
            app_label='sources', model_name='Source'
        )

        form_list = DocumentCreateWizardStep.get_choices(
            attribute_name='form_class'
        )
        condition_dict = dict(
            DocumentCreateWizardStep.get_choices(attribute_name='condition')
        )

        result = self.__class__.get_initkwargs(
            condition_dict=condition_dict, form_list=form_list
        )
        self.form_list = result['form_list']
        self.condition_dict = result['condition_dict']

        for source in Source.objects.filter(enabled=True):
            try:
                action = source.get_action(name='document_upload')
                if action.has_interface(interface_name='View'):
                    return super().dispatch(request=request, *args, **kwargs)
            except SourceActionExceptionUnknown:
                """
                Non fatal. Ignore and try the next source.
                """

        messages.error(
            message=_(
                message='No interactive document sources have been defined or '
                'none have been enabled, create one before proceeding.'
            ), request=request
        )
        return HttpResponseRedirect(
            redirect_to=reverse(viewname='sources:source_list')
        )

    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(form=form, **kwargs)

        wizard_step = DocumentCreateWizardStep.get(name=self.steps.current)

        context.update(
            {
                'form_css_classes': 'form-hotkey-enter form-hotkey-double-click',
                'step_title': _(
                    message='Step %(step)d of %(total_steps)d: %(step_label)s'
                ) % {
                    'step': self.steps.step1,
                    'step_label': wizard_step.label,
                    'total_steps': len(self.form_list)
                },
                'title': _(message='Document upload wizard'),
                'wizard_step': wizard_step,
                'wizard_steps': DocumentCreateWizardStep.get_all()
            }
        )

        context['form_button_overrides'] = (
            {
                'icon': icon_wizard_step_first,
                'label': _(message='First'),
                'name_override': 'wizard_goto_step',
                'value': self.steps.first
            },
            {
                'icon': icon_wizard_step_previous,
                'label': _(message='Previous'),
                'name_override': 'wizard_goto_step',
                'value': self.steps.prev
            },
            {
                'icon': icon_wizard_step_next,
                'is_primary': True,
                'label': _(message='Next')
            }
        )

        if not self.steps.prev:
            context['form_button_overrides'][0]['css_classes'] = 'disabled'
            context['form_button_overrides'][0]['disabled'] = True
            context['form_button_overrides'][1]['css_classes'] = 'disabled'
            context['form_button_overrides'][1]['disabled'] = True

        if not wizard_step.get_next_is_enabled(wizard=self):
            context['form_button_overrides'][2]['css_classes'] = 'disabled'
            context['form_button_overrides'][2]['disabled'] = True

        return context

    def get_form_initial(self, step):
        wizard_step = DocumentCreateWizardStep.get(name=step)

        initial = wizard_step.get_form_initial(wizard=self) or {}

        return initial

    def get_form_kwargs(self, step):
        wizard_step = DocumentCreateWizardStep.get(name=step)

        kwargs = wizard_step.get_form_kwargs(wizard=self) or {}

        return kwargs

    def done(self, form_list, **kwargs):
        query_dict = {}

        for step in DocumentCreateWizardStep.get_all():
            query_dict.update(
                step.done(wizard=self) or {}
            )

        url = furl(
            reverse(viewname='sources:document_upload')
        )
        # Use equal and not .update() to get the same result as using
        # urlencode(doseq=True).
        url.args = query_dict

        return HttpResponseRedirect(
            redirect_to=url.tostr()
        )



from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.model_mixins import BackendModelMixin
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .classes import MailerBackendNull
from .events import (
    event_mailing_profile_created, event_mailing_profile_edited
)
from .managers import UserMailerManager
from .model_mixins import UserMailerBusinessLogicMixin


class UserMailer(
    BackendModelMixin, ExtraDataModelMixin, UserMailerBusinessLogicMixin,
    models.Model
):
    """
    This model is used to create mailing profiles that can be used from
    inside the system. These profiles differ from the system mailing
    profile in that they can be created at runtime and can be assigned
    ACLs to restrict their use.
    """
    _backend_model_null_backend = MailerBackendNull
    _ordering_fields = ('default', 'enabled', 'label')

    label = models.CharField(
        help_text=_(message='A short text describing the mailing profile.'),
        max_length=128, unique=True, verbose_name=_(message='Label')
    )
    default = models.BooleanField(
        default=True, help_text=_(
            message='If default, this mailing profile will be '
            'pre-selected on the document mailing form.'
        ), verbose_name=_(message='Default')
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_(message='Enabled')
    )

    objects = UserMailerManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _(message='Mailing profile')
        verbose_name_plural = _(message='Mailing profiles')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(viewname='mailer:mailing_profile_list')

    def natural_key(self):
        return (self.label,)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_mailing_profile_created,
            'target': 'self'
        },
        edited={
            'event': event_mailing_profile_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        if self.default:
            UserMailer.objects.select_for_update().exclude(
                pk=self.pk
            ).update(default=False)

        return super().save(*args, **kwargs)



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from mayan.apps.common.utils import get_class_full_name

from .classes import (
    AuthenticationBackend, AuthenticationBackendRememberMeMixin
)
from .django_authentication_backends import (
    DjangoAuthenticationBackendModelEmail
)
from .forms import (
    AuthenticationFormEmailPassword, AuthenticationFormUsernamePassword
)
from .literals import SESSION_MULTI_FACTOR_USER_ID_KEY


class AuthenticationBackendModelDjangoDefault(
    AuthenticationBackendRememberMeMixin, AuthenticationBackend
):
    """Bare bones backend that preserves Django's default behaviors."""


class AuthenticationBackendModelUsernamePassword(
    AuthenticationBackendRememberMeMixin, AuthenticationBackend
):
    login_form_class = AuthenticationFormUsernamePassword

    def get_user(self, request, form_list=None, kwargs=None):
        return get_user_model().objects.get(
            pk=request.session[SESSION_MULTI_FACTOR_USER_ID_KEY]
        )

    def initialize(self):
        settings.AUTHENTICATION_BACKENDS = (
            get_class_full_name(klass=ModelBackend),
        )


class AuthenticationBackendModelEmailPassword(
    AuthenticationBackendRememberMeMixin, AuthenticationBackend
):
    login_form_class = AuthenticationFormEmailPassword

    def get_user(self, request, form_list=None, kwargs=None):
        return get_user_model().objects.get(
            pk=request.session[SESSION_MULTI_FACTOR_USER_ID_KEY]
        )

    def initialize(self):
        settings.AUTHENTICATION_BACKENDS = (
            get_class_full_name(klass=DjangoAuthenticationBackendModelEmail),
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Logging'), name='logging'
)

event_error_log_deleted = namespace.add_event_type(
    label=_(message='Error log deleted'), name='error_log_deleted'
)



from django.apps import apps

from mayan.apps.common.classes import PropertyHelper


class DocumentMetadataHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentMetadataHelper(*args, **kwargs)

    def get_result(self, name):
        DocumentMetadata = apps.get_model(
            app_label='metadata', model_name='DocumentMetadata'
        )

        try:
            document_metadata = self.instance.metadata.get(
                metadata_type__name=name
            )
        except DocumentMetadata.DoesNotExist:
            return None
        else:
            return document_metadata.value



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_VIEWS_PAGINATE_BY, DEFAULT_VIEWS_PAGING_ARGUMENT
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Views'), name='views'
)

setting_paginate_by = setting_namespace.do_setting_add(
    default=DEFAULT_VIEWS_PAGINATE_BY, global_name='VIEWS_PAGINATE_BY',
    help_text=_(
        message='The number objects that will be displayed per page.'
    )
)
setting_paging_argument = setting_namespace.do_setting_add(
    default=DEFAULT_VIEWS_PAGING_ARGUMENT,
    global_name='VIEWS_PAGING_ARGUMENT', help_text=_(
        message='A string specifying the name to use for the paging parameter.'
    )
)
